import fmodpy

# Import the profiler.
profiler = fmodpy.fimport("../axy_profiler.f90", verbose=False, build_dir="test_axy_profiler", name="test_axy_profiler").profiler

# Run a simple test with a sleep command.
def test_profiler():
    print("test profiler ", end="...", flush=True)
    import time
    profiler.start_profiling("time_test")
    time.sleep(1)
    profiler.stop_profiling("time_test")
    prof = profiler.get_profile("time_test")
    assert (prof.wall_time > 1)
    assert (prof.cpu_time > 0)
    print(" passed.", flush=True)

if __name__ == "__main__":
    test_profiler()



import numpy as np

# This subclass of a NumPy matrix has relevant information to
# the process used when transforming a Data object into a
# purely numerical matrix of values for prediction purposes.
# 
# Numeric.to_real   -- the function for processing regular
#                      vectors into associated real vectors
# Numeric.from_real -- the function for processing real vectors
#                      back into regular vector
# Numeric.in_names  -- names of columns before being mapped
# Numeric.in_types  -- types of columns before being mapped
# Numeric.mappings  -- the mappings of input column values to outputs
# Numeric.groups    -- the collections of real columns tied to input columns
# Numeric.dropped   -- columns that are dropped from input
# Numeric.names     -- meaningful column names for the real vectors
# Numeric.nums      -- indices of numerical value outputs
# Numeric.cats      -- indices of categorical mapping outputs
# Numeric.shift     -- How to shift this data to have min 0.
# Numeric.scale     -- How to scale this data (after shift) to 
#                      have a domain of [0., 1.]                    
class Numeric(np.ndarray):

    # ----------------------------------------------------------------
    # Initialize this Numeric object with special Numpy ndarray
    # subclassing methods (these are needed for correct behavior).
    def __new__(cls, data, in_names, in_types, mappings, groups,
                dropped, names, nums, cats, print_column_width):
        # Either an actual matrix or a shape can be provided.
        if (type(data) is tuple):
            self = np.asarray(np.zeros(data)).view(cls)
        else:
            self = np.asarray(data).view(cls)
        # Make sure the shape is correct.
        assert (len(names) == self.shape[1]), f"Data number of columns ({self.shape[1]}) must equal length of names ({len(names)})."
        # Assign extra local information.
        self.in_names = in_names
        self.in_types = in_types
        self.groups = groups
        self.mappings = mappings
        self.dropped = dropped
        self.names = names
        self.nums = nums
        self.cats = cats
        self.print_column_width = print_column_width
        # Store an internal lookup table that maps back to original values.
        self.np_mappings = {}
        for col in mappings:
            values = []
            encodings = []
            for (val,encoding) in mappings[col].items():
                values.append(val)
                encodings.append(encoding)
            self.np_mappings[col] = (np.asarray(encodings), values)
        # Compute the shift and scale for normalization (of numeric values)
        self.shift = np.zeros(self.shape[1])
        self.scale = np.ones(self.shape[1])
        if (type(data) is not tuple):
            self.compute_shift_and_scale()
        return self
    def __array_finalize__(self, obj):
        if obj is None: return
        # Assign extra local information.
        self.in_names = getattr(obj, "in_names", None)
        self.in_types = getattr(obj, "in_types", None)
        self.groups = getattr(obj, "groups", None)
        self.mappings = getattr(obj, "mappings", None)
        self.dropped = getattr(obj, "dropped", None)
        self.names = getattr(obj, "names", None)
        self.nums = getattr(obj, "nums", None)
        self.cats = getattr(obj, "cats", None)
        self.print_column_width = getattr(obj, "print_column_width", None)
        self.np_mappings = getattr(obj, "np_mappings", None)
        # Compute the shift and scale for normalization (of numeric values)
        self.shift = getattr(obj, "shift", None)
        self.scale = getattr(obj, "scale", None)
    # Define a reduce function that includes internal attributes.
    def __reduce__(self):
        cls, internal, state = super().__reduce__()
        state = state + (
            self.in_names, self.in_types, self.groups, self.mappings,
            self.dropped, self.names, self.nums, self.cats,
            self.print_column_width, self.np_mappings, self.shift, self.scale)
        return (cls, internal, state)
    # Define a setstate function that captures internal attributes.
    def __setstate__(self, state):
        np_state, custom_state = state[:-12], state[-12:]
        self.in_names, self.in_types, self.groups, self.mappings, \
            self.dropped, self.names, self.nums, self.cats, \
            self.print_column_width, self.np_mappings, self.shift, \
            self.scale = custom_state
        super().__setstate__(np_state)
    # ----------------------------------------------------------------

    # Update the calculation of the shift and scale for this Numeric object.
    def compute_shift_and_scale(self):
        self.shift[self.nums] = np.min(self[:,self.nums], axis=0)
        self.scale[self.nums] = np.max(self[:,self.nums], axis=0) - self.shift[self.nums]

    # Takes a row in the original space and generates a row in the
    # associated real-valued space.
    def to_real(self, row, fill=False):
        # Check for row length
        if (len(row) != len(self.in_names)):
            raise(Data.BadElement(f"Provided row has {len(row)} elements, expected {len(names)}."))
        # Bulid the real-valued vector
        real_row = []
        for i,(n,t,v) in enumerate(zip(self.in_names, self.in_types, row)):
            # Skip dropped columns.
            if (i in self.dropped): continue
            # If there is no map, then we must exit here.
            if (n not in self.mappings):
                from .exceptions import BadValue
                raise(Data.BadValue(f"No known mapping for column '{n}' to real values."))
            n_map = self.mappings[n]
            # Check if this incoming value is missing.
            if (v is None):
                if (None not in n_map):
                    from .exceptions import BadValue
                    raise(BadValue(f"Value '{v}' for column '{n}' was unexpected and cannot be mapped to real."))
                real_row += n_map[None]
            # Check if the type if float or int (literally translate).
            elif (t is float):
                real_row += [v]
            elif (t is int):
                real_row += [float(v)]
            # Otherwise, this value needs to be mapped.
            else:
                # If the value is not known..
                if (v not in n_map):
                    from .exceptions import BadValue
                    raise(BadValue(f"Value '{v}' for column '{n}' has no known real encoding."))
                real_row += n_map[v]
        # Check for row length
        if (len(real_row) != len(self.names)):
            from .exceptions import BadElement
            raise(BadElement(f"Provided row has {len(row)} elements, translation to {len(self.names)} unsuccesful, only created {len(real_row)}."))
        # Return the real-valued vector
        return real_row


    # Takes a row in the real-valued space and generates a row in the
    # original space.
    def from_real(self, real_row, bias={}):
        import numpy as np
        # Verify length of real-vector
        real_row = list(real_row)
        if (len(real_row) != len(self.names)):
            from .exceptions import BadElement
            raise(BadElement(f"Provided real vector has {len(real_row)} elements, expected {len(self.names)}."))
        # Build the original row
        row = []
        for (n,t) in zip(self.in_names, self.in_types):
            cols = self.groups[n]
            if (t is float):
                row += [real_row[cols[0]]]
            elif (t is int):
                row += [int(round(real_row[cols[0]]))]
            else:
                encoding = np.asarray([real_row[c] for c in cols])
                encodings, values = self.np_mappings[n]
                nearest = np.argmin(np.linalg.norm(encodings - encoding, axis=1))
                row += [values[nearest]]
        # Check for row length
        if (len(row) != len(self.in_names)):
            from .exceptions import BadElement
            raise(BadElement(f"Provided row has {len(real_row)} elements, translation to {len(self.in_names)} unsuccesful, only created {len(row)}."))
        return row


    # Produce a string that includes semantic information about this object.
    def __str__(self):
        opener = "  ["
        str_real_names = [opener]
        for n in self.names:
            if not ((len(str_real_names[0]) == len(opener)) or
                    (len(str_real_names[-1]) + len(n) < self.print_column_width)):
                str_real_names.append("   ")
            str_real_names[-1] += f"'{n}', "
        str_real_names[-1] = str_real_names[-1][:-2] + "]"
        str_real_names = "\n".join(str_real_names)
        return f"Numeric object for data with {len(self.names)} columns named:\n{str_real_names}\n\n{super().__str__()}"


# import inspect
# # Assign the comments to the official documentation spot.
# Numeric.__doc__ = inspect.getcomments(Numeric).replace("# ","")




def worker(arr):
    # Access and modify the shared array
    shared_arr = np.ctypeslib.as_array(arr)
    shared_arr[0] = 42

def _test_sharedctypes():
    import numpy as np
    from multiprocessing import Process, sharedctypes
    from ctypes import c_double
    from tlux.profiling import mem_use
    print("mem_use(): ", mem_use(), flush=True)
    # Create a large NumPy array
    initial_mem_use = mem_use()
    large_array = np.zeros((1000, 1000))
    new_mem_use = mem_use()
    array_size = new_mem_use - initial_mem_use
    print("array_size: ", array_size, flush=True)
    # Create a shared memory object
    initial_mem_use = mem_use()
    shared_arr = sharedctypes.RawArray(c_double, large_array.flatten())
    new_mem_use = mem_use()
    shared_array_size = new_mem_use - initial_mem_use
    print("shared_array_size: ", shared_array_size, flush=True)
    # Convert the shared memory object to a NumPy array
    initial_mem_use = mem_use()
    shared_np_array = np.frombuffer(shared_arr, dtype=np.float64).reshape(large_array.shape)
    new_mem_use = mem_use()
    np_buffer_size = new_mem_use - initial_mem_use
    print("np_buffer_size: ", np_buffer_size, flush=True)
    # Start a worker process
    initial_mem_use = mem_use()
    p = Process(target=worker, args=(shared_arr,))
    p.start()
    new_mem_use = mem_use()
    process_size = new_mem_use - initial_mem_use
    p.join()
    print("process_size: ", process_size, flush=True)
    # Check if the shared array was modified
    assert (shared_np_array[0,0] == 42)
    assert (shared_np_array[0,1:].sum() == 0.0)


class SlowObject:
    value = 0.1
    def wait(self, value=None):
        if value is None: value = self.value
        import time
        time.sleep(value)
        return value


def _test_asynchronous(verbose=False, error_tolerance = 0.001):
    print(" test Asynchronous..", end="\n" if verbose else "", flush=True)
    # 
    from tlux.system import Asynchronous, Timer
    AsyncSlowObject = Asynchronous(SlowObject, verbose=False)
    t = Timer()
    # Initialization.
    obj = AsyncSlowObject()
    if verbose: print("  type(obj): ", type(obj), t(), flush=True)
    # Try to look up a missing attribute.
    try: obj.does_not_exist
    except AttributeError: pass
    else: raise(RuntimeError(f"Expected 'AttributeError' when looking up 'obj.does_not_exist'."))
    # Attribute lookup.
    t.start()
    value = obj.value
    assert (t() < error_tolerance), f"Object took too long to fetch an attribute, {t()} seconds."
    if verbose: print("  obj.value: ", value, t(), flush=True)
    # Attribute value retrieval.
    t.start()
    value = obj.value.get()
    assert (t() < error_tolerance), f"Object took too long to get an attribute's value, {t()} seconds."
    if verbose: print("  obj.value.get(): ", value, t(), flush=True)
    # Expensive method call.
    t.start()
    value = obj.wait()
    assert (t() < error_tolerance), f"Object took too long to start working on a method, {t()} seconds."
    if verbose: print("  type(obj.wait()): ", type(value), t(), flush=True)
    # Waiting on expensive method results.
    value = value.get()
    assert (value == SlowObject.value), f"Returned waited time \"{value}\" does not match expected default {SlowObject.value}."
    assert (t() >= SlowObject.value-error_tolerance), f"Object recieved result faster than the expected ~{SlowObject.value} seconds, {t()} seconds."
    if verbose: print("  obj.wait().get(): ", value, t(), flush=True)
    # Try triggering an error within the Aynchronous instance.
    value = obj.wait('hello').get()
    assert (type(value) == TypeError), f"Expected to receive a 'TypeError', but got a '{type(value).__name__}'."
    assert (str(value) == "'str' object cannot be interpreted as an integer"), f"Expected a specific error message, but did not receive the expected message."
    # 
    print(" passed.", flush=True)


if __name__ == "__main__":
    # _test_sharedctypes()
    _test_asynchronous()



import os, random
import numpy as np
import ctypes
from tlux.approximate.axy.preprocessing import i_map, i_encode, to_array
from tlux.unique import unique, to_int
from tlux.profiling import profile



def _test_imap(num_words=10, max_word_length=16, leave_save=False, verbose=False):
    if (not verbose): print("test imap", ("load file " if leave_save == False else "save file "), end="...", flush=True)
    random.seed(0)
    path = f"words-{max_word_length}_{num_words}.npz"
    if (not os.path.exists(path)):
        if verbose: print(f"Generating {num_words} words..", flush=True)
        # Generate random words.
        words = [[
            "".join((chr(random.randint(ord('0'),ord('z')))
                     for _ in range(random.randint(0,max_word_length))))
              .replace("\\", "/")
              .replace("`", ".")
            , random.randint(0, max_word_length**2)] for _ in range(num_words)
        ]
        # Convert those words to a numpy array.
        xi = np.asarray(words, dtype=object)
        # Save the words to file (for later reuse).
        np.savez_compressed(path, xi=xi)
    else:
        if verbose: print(f"Loading '{path}'..", flush=True)
        xi = np.load(path, allow_pickle=True)['xi']

    # Delete the save file if "leave_save" is false.
    if (not leave_save):
        os.remove(path)

    # Modify the shape.
    if verbose: print("Modifying..", flush=True)
    xi = xi.reshape((-1,2))

    # @profile
    # def test(steps=3):
    #     for _ in range(steps):
    #         print("", _, end="\r")
    #         res = unique(xi)
    #         del(res)
    # for _ in range(10):
    #     test()
    #     test.show_profile()
    # print()
    # unique.show_profile()
    # exit()

    # Convert data into the xi_map.
    if verbose: print("Calling i_map..", flush=True)
    xi_map = i_map(xi, [])
    if verbose:
        if (hasattr(unique, "show_profile")): unique.show_profile()
        if (hasattr(i_map, "show_profile")): i_map.show_profile()
        for i,m in enumerate(xi_map):
            print(f" map {i} (len {len(m)})", m)

    # Assert that the maps have the expected contents in the expected order.
    #   order first by length, second by ordinal value
    assert (tuple(xi_map[0]) == ('@C', 'ub', 'VvU', ']gX', 'HG4Ql', 'Bui;:Xq', 'A<PtBW<9Z', 'uJvTh;aXyN', 'e5QqncVm]zKp', 'hrQ7v1;c0oZOY8H'))
    assert (tuple(xi_map[1]) == ('19', '35', '63', '71', '104', '113', '141', '148', '241', '250'))

    # Encode the data in integer format.
    if verbose: print("Calling i_encode..", flush=True)
    final_xi = i_encode(xi, xi_map)
    if verbose:
        if (hasattr(to_int, "show_profile")): to_int.show_profile()
        if (hasattr(i_encode, "show_profile")): i_encode.show_profile()
        # Print the source data and the final integer version.
        print(xi)
        print(final_xi)

    assert (final_xi == np.asarray(
        [[9, 14],
         [7, 19],
         [4, 15],
         [10, 16],
         [6, 20],
         [3, 13],
         [8, 18],
         [5, 12],
         [1, 11],
         [2, 17]]
    )).all()

    if (not verbose): print(" passed.", flush=True)

if __name__ == "__main__":
    _test_imap(verbose=False, leave_save=True)
    _test_imap(verbose=False, leave_save=False)



# This file is (an attempt at a) reimplementation of the core Fortran library
#  routines for the AXY model entirely in python. This is much slower, but
#  provides the portability and readability benefits of being in python.
# 
# This implementation is occasionally lagging, so be cautious with its usage
#  and prepared to submit an Issue in the case it produces incongruent outputs.


import ctypes
import numpy as np
from tlux.approximate.axy.summary import AxyModel


# Cast the elements of an array to the specified data type.
#
# The function accepts an array of elements, a desired data type `dtype`, and an optional
# parameter `order` that specifies the memory layout of the resulting array. It attempts to
# cast each element of the array to the specified data type, preserving the shape of the input array.
# If `dtype` is passed as a type, then each element is cast to that type before creating the array.
# Otherwise, `dtype` should be a valid data type string recognizable by numpy.
#
# Parameters
# ----------
# arr : array_like
#     The input array whose elements are to be cast to the desired data type.
# dtype : type or str
#     The desired data type to which the elements of the input array are to be cast. 
#     Can be a type (e.g., int, float) or a string that numpy recognizes (e.g., 'int32', 'float64').
# order : {'C', 'F'}, optional, default: 'F'
#     Specify the memory layout of the output array. If 'C', the array will be in C-contiguous
#     order (row-major). If 'F', the array will be in Fortran-contiguous order (column-major).
#
# Returns
# -------
# ndarray
#     A numpy array with the same shape as the input array but with elements cast to the specified data type.
#
# Examples
# --------
# >>> cast([[1.2, 2.3], [3.4, 4.5]], int)
# array([[1, 2], [3, 4]])
# >>> cast([1, 2, 3], 'float64')
# array([1., 2., 3.])
# >>> cast([[1, 2], [3, 4]], complex, order='C')
# array([[1.+0.j, 2.+0.j], [3.+0.j, 4.+0.j]])
#
# Raises
# ------
# ValueError
#     If `dtype` is not recognized as a valid data type.
# TypeError
#     If `arr` is not array_like.
#
def cast(arr, dtype, order="F"):
    arr = np.asarray(arr)
    if (type(dtype) is type):
        arr = np.asarray(
            [dtype(v) for v in arr.flatten()],
            dtype=dtype,
            order=order
        ).reshape(arr.shape)
    else:
        arr = np.asarray(arr, dtype=dtype, order=order)
    return arr


# Convert a single index to a pair of indices.
# 
# Parameters
# ----------
# n : int
#     The number of source elements / the largest number in any pair.
# i : int
#     The single index to be converted to a pair of indices that
#     is 1-indexed and in the range [1, num_elements**2].
# 
# Returns
# -------
# tuple[int, int]
#   A pair of indices (pair1, pair2) that come from a function
#    the generates an ordering such that all pairs involving
#    smaller numbers come first.
# 
# Example
# -------
#  >>> [index_to_pair(3, i) for i in range(1,10)]
#  [(1,1) (1,2) (2,1) (1,3) (3,1) (2,2) (2,3) (3,2) (3,3)]
# 
def index_to_pair(num_elements, i):
    index_from_back = num_elements**2 - i
    group_from_back = int(index_from_back**(1/2))
    group = num_elements - group_from_back
    remaining_in_group = index_from_back - group_from_back**2
    group_size = 2*group_from_back + 1
    index_in_group = group_size - remaining_in_group
    other = group + index_in_group // 2
    if (index_in_group % 2 == 0):
        return (group, other)
    else:
        return (other, group)


# Check if the shapes and values of the input arrays match the expectations based on the model configuration.
# 
# Parameters
# ----------
# config : ModelConfig
#     Model configuration object with attributes:
#     - TOTAL_SIZE: expected total size of the model array.
#     - MDN: expected first dimension size of the X input array.
#     - MDO: expected first dimension size of the Y output array when greater than 0.
#     - ADO: expected first dimension size of the Y output array when MDO is 0.
#     - MNE: expected maximum value in the XI input array when greater than 0.
#     - ADI: flag that when greater than 0, SIZES array size must match Y second dimension size.
#     - PARTIAL_AGGREGATION: flag to indicate whether partial aggregation is allowed.
#     - ADN: expected first dimension size of the AX input array.
#     - ANE: expected maximum value in the AXI input array.
# 
# model : ndarray of shape (n,)
#     Model array of real numbers.
# 
# ax : ndarray of shape (adn, m)
#     AX input array of real numbers.
# 
# axi : ndarray of shape (p, m)
#     AXI input array of integers.
# 
# sizes : ndarray of shape (q,)
#     Sizes array of integers.
# 
# x : ndarray of shape (mdn, k)
#     X input array of real numbers.
# 
# xi : ndarray of shape (r, k)
#     XI input array of integers.
# 
# y : ndarray of shape (s, k)
#     Y input array of real numbers.
# 
# Returns
# -------
# info : int
#     Returns nonzero INFO if any shapes or values do not match expectations:
#     1: Model size does not match model configuration.
#     2: Input arrays do not match in size.
#     3: X input dimension is bad.
#     4: Model output dimension is bad, does not match Y.
#     5: Aggregator output dimension is bad, does not match Y.
#     6: Input integer XI size does not match X.
#     7: Input integer X index out of range.
#     8: SIZES has wrong size.
#     9: AX and SUM(SIZES) do not match.
#     10: AX input dimension is bad.
#     11: Input integer AXI size does not match AX.
#     12: Input integer AX index out of range.
# 
def check_shape(config, model, ax, axi, sizes, x, xi, y):
    # Compute whether the shape matches the CONFIG
    if (model.size != config.TOTAL_SIZE):
        return 1  # Model size does not match model configuration.
    if (x.shape[1] != y.shape[1]):
        return 2  # Input arrays do not match in size.
    if (x.shape[0] != config.MDN):
        return 3  # X input dimension is bad.
    if ((config.MDO > 0) and (y.shape[0] != config.MDO)):
        return 4  # Model output dimension is bad, does not match Y.
    if ((config.MDO == 0) and (y.shape[0] != config.ADO)):
        return 5  # Aggregator output dimension is bad, does not match Y.
    if ((config.MNE > 0) and (xi.shape[1] != x.shape[1])):
        return 6  # Input integer XI size does not match X.
    if ((xi.min() < 0) or (xi.max() > config.MNE)):
        return 7  # Input integer X index out of range.
    if ((config.ADI > 0) and (not config.PARTIAL_AGGREGATION) and (sizes.size != y.shape[1])):
        return 8  # SIZES has wrong size.
    if (ax.shape[1] != sizes.sum()):
        return 9  # AX and SUM(SIZES) do not match.
    if (ax.shape[0] != config.ADN):
        return 10  # AX input dimension is bad.
    if (axi.shape[1] != ax.shape[1]):
        return 11  # Input integer AXI size does not match AX.
    if ((axi.min() < 0) or (axi.max() > config.ANE)):
        return 12  # Input integer AX index out of range.
    return 0  # All checks passed.



# Normalize numeric input values.
# 
# Parameters
# ----------
# config : object
#     An object representing the model configuration, expected to have the following attributes:
#         NEEDS_SHIFTING : bool
#         NEEDS_SCALING : bool
#         ADN : int
#         MDN : int
#         AISS, AISE, MISS, MISE : int
#         AIMS, AIME, MIMS, MIME : int
#         ADO, MDO : int
#         NORMALIZE : bool
#         NEEDS_CLEANING : bool
# 
# model : ndarray
#     1D array of shape (n,) representing the model.
# 
# ax : ndarray
#     2D array of shape (m, p) representing aggregate inputs.
# 
# sizes : ndarray
#     1D array of shape (s,) representing sizes.
# 
# x : ndarray
#     2D array of shape (q, r) representing fixed inputs.
# 
# Returns
# -------
# ax : ndarray
#     2D array of normalized aggregate inputs.
# 
# x : ndarray
#     2D array of normalized fixed inputs.
# 
def normalize_inputs(config, model, ax, sizes, x):
    # If normalization needs to happen..
    if config.NORMALIZE:
        # Normalize the aggregate inputs.
        if (config.ADO > 0) and (config.ADN > 0):
            # Apply shift terms to aggregator inputs.
            if config.NEEDS_SHIFTING:
                ax_shift = model[config.AISS:config.AISE]
                ax[:config.ADN,:] += ax_shift[:, np.newaxis]

            # Remove any NaN or Inf values from the data.
            if config.NEEDS_CLEANING:
                ax[:config.ADN,:][np.isnan(ax[:config.ADN,:]) | ~np.isfinite(ax[:config.ADN,:])] = 0.0

            # Apply multiplier.
            if config.NEEDS_SCALING:
                ax_rescale = model[config.AIMS:config.AIME].reshape(config.ADN, config.ADN)
                ax[:config.ADN,:] = np.dot(ax_rescale.T, ax[:config.ADN,:])

        # Normalize the fixed inputs.
        if (config.MDO > 0) and (config.MDN > 0):
            # Apply shift terms to numeric model inputs.
            if config.NEEDS_SHIFTING:
                x_shift = model[config.MISS:config.MISE]
                x[:config.MDN,:] += x_shift[:, np.newaxis]

            # Remove any NaN or Inf values from the data.
            if config.NEEDS_CLEANING:
                x[:config.MDN,:][np.isnan(x[:config.MDN,:]) | ~np.isfinite(x[:config.MDN,:])] = 0.0

            # Apply multiplier.
            if config.NEEDS_SCALING:
                x_rescale = model[config.MIMS:config.MIME].reshape(config.MDN, config.MDN)
                x[:config.MDN,:] = np.dot(x_rescale.T, x[:config.MDN,:])

    return ax, x, 0




# Given the raw input data, fetch a new set of data that fits in memory.
# 
# Parameters
# ----------
# config : object
#     Config object, contains model configuration.
# agg_iterators_in : np.ndarray
#     2D numpy array of int64.
# ax_in : np.ndarray
#     2D numpy array of float.
# ax : np.ndarray
#     2D numpy array of float.
# axi_in : np.ndarray
#     2D numpy array of int64.
# axi : np.ndarray
#     2D numpy array of int64.
# sizes_in : np.ndarray
#     1D numpy array of int64.
# sizes : np.ndarray
#     1D numpy array of int64.
# x_in : np.ndarray
#     2D numpy array of float.
# x : np.ndarray
#     2D numpy array of float.
# xi_in : np.ndarray
#     2D numpy array of int64.
# xi : np.ndarray
#     2D numpy array of int64.
# y_in : np.ndarray
#     2D numpy array of float.
# y : np.ndarray
#     2D numpy array of float.
# yw_in : np.ndarray
#     2D numpy array of float.
# yw : np.ndarray
#     2D numpy array of float.
# 
# Returns
# -------
# na : int
#     Total number of aggregate inputs that were added.
# nm : int
#     Total number of inputs.
# 
def fetch_data(config, agg_iterators_in, ax_in, ax, axi_in, axi, sizes_in, sizes, x_in, x, xi_in, xi, y_in, y, yw_in, yw):
    # Local variables
    one = np.int64(1)
    zero = np.int64(0)
    rt = np.float64

    # Allocate storage space if it will be used.
    if sizes_in.size > 0:
        rsizes = np.empty(config.nm, dtype=rt)
        sorted_order = np.empty(config.nm, dtype=np.int64)

    # Pack the regular inputs into the working space, storing sizes.
    i_next = config.i_next
    i_mult = config.i_mult
    i_step = config.i_step
    i_iter = config.i_iter

    for i in range(1, min(config.nm, y_in.shape[1]) + 1):
        # Choose iteration strategy
        if config.nm >= y_in.shape[1] and not config.partial_aggregation:
            gendex = i
        else:
            gendex = get_next_index(config.nmt, config.i_next, config.i_mult,
                                    config.i_step, config.i_mod, config.i_iter)
        # Store the size
        if sizes_in.size > 0:
            rsizes[i - 1] = rt(sizes_in[gendex - 1])
            if config.pairwise_aggregation:
                rsizes[i - 1] = rsizes[i - 1]**2

        x[:config.mdn, i - 1] = x_in[:, gendex - 1]
        xi[:, i - 1] = xi_in[:, gendex - 1]
        y[:, i - 1] = y_in[:, gendex - 1]
        yw[:, i - 1] = yw_in[:, gendex - 1]

    # TODO: Get the aggregate inputs too.



# Define the full EVALUATE function in python (reimplementation).
def evaluate(config, model, ax, axi, sizes, x, xi, dtype="float32", **unused_kwargs):
    # Get some constants.
    nm = x.shape[1]
    na = ax.shape[1]
    m = AxyModel(config, cast(model, dtype))
    state_values = {}
    # Embed the AXI values.
    ax_embedded = cast(np.zeros((config.adi, na)), dtype)
    for n in range(axi.shape[1]):
        ax_embedded[:config.adn,n] = ax[:config.adn,n]
        for d in range(axi.shape[0]):
            e = axi[d,n]
            if (e > 0) and (e <= config.ane):
                ax_embedded[-config.ade:,n] += m.a_embeddings[:,e-1]
            elif (e > config.ane):
                e1, e2 = index_to_pair(num_elements=config.ane+1, i=e-config.ane)
                ax_embedded[-config.ade:,n] += m.a_embeddings[:,e1-1-1] - m.a_embeddings[:,e2-1-1]
        if (axi.shape[0] > 1):
            ax_embedded[-config.ade:,n] /= axi.shape[0]
    ax = ax_embedded
    state_values["ax"] = ax
    # Embed the XI values.
    x_embedded = cast(np.zeros((config.mdi, nm)), dtype)
    for n in range(xi.shape[1]):
        x_embedded[:config.mdn,n] = x[:config.mdn,n]
        for d in range(xi.shape[0]):
            e = xi[d,n]
            if (e > 0) and (e <= config.mne):
                x_embedded[config.mdn:config.mdn+config.mde:,n] += m.m_embeddings[:,e-1]
            elif (e > config.mne):
                e1, e2 = index_to_pair(num_elements=config.mne, i=e)
                x_embedded[config.mdn:config.mdn+config.mde,n] += m.m_embeddings[:,e1-1] - m.m_embeddings[:,e2-1]
        if (xi.shape[0] > 1):
            x_embedded[-config.mde:,n] /= xi.shape[0]
    x = x_embedded
    state_values["x"] = x
    # Initialize a holder for the output.
    y = cast(np.zeros((config.do, nm)), dtype)
    # Evaluate the aggregator.
    if (config.ado > 0):
        if (config.normalize and (config.adn > 0)):
            # Add the shift term (across all point column vectors).
            ax[:config.adn,:] = (ax[:config.adn,:].T + m.ax_shift).T
            # Replace all NaN or Inf values
            ax[:config.adn,:] = np.where(
                np.logical_or(
                    np.isnan(ax[:config.adn,:].astype("float32")),
                    np.isinf(ax[:config.adn,:].astype("float32")),
                ),
                0.0,
                ax[:config.adn,:]
            )
            # Apply the input multiplier.
            if (config.needs_scaling):
                ax[:config.adn,:] = m.ax_rescale.T @ ax[:config.adn,:]
        # Evaluate the MLP.
        values = ax
        state_values["a_states"] = []
        if (config.ans > 0):
            # Apply input transformation.
            values = np.clip(
                ((values.T @ m.a_input_vecs) + m.a_input_shift).T,
                config.discontinuity, float('inf')
            )
            state_values["a_states"].append(values)
            # Apply internal transformations.
            for i in range(config.ans-1):
                values = np.clip(
                    ((values.T @ m.a_state_vecs[:,:,i]) + m.a_state_shift[:,i]).T,
                    config.discontinuity, float('inf')
                )
                state_values["a_states"].append(values)
        state_values["a_states"] = np.asarray(state_values["a_states"]).T
        # Apply output transformation.
        ay = (values.T @ m.a_output_vecs[:,:])
        state_values["ay"] = ay
        ay_error = ay[:,config.ado:config.ado+1]  # extract +1 for error prediction
        ay = ay[:,:config.ado] # strip off +1 for error prediction
        # If there is a following model..
        if (config.mdo > 0):
            # Apply output shift.
            ay[:,:] = (ay + m.ay_shift)
            # Compute the first aggregator output embedding position.
            e = config.mdn + config.mde
            # Set the aggregator output to be a slice of X.
            agg_out = x[e:,:]
        else:
            # Set the aggregator output to be Y.
            agg_out = y[:,:]
        # Aggregate the batches. With partial aggregation, we have one output for
        #  partial mean starting from the "last" aggregate output.
        if (config.partial_aggregation):
            f_start = 0
            a_start = 0
            for i,s in enumerate(sizes):
                a_end = a_start + s
                f_end = f_start + max(1,s)
                for out_i, agg_i in zip(range(f_end-1, f_start-1, -1), range(a_end-1, a_start-1, -1)):
                    num_elements = a_end - agg_i
                    agg_out[:,out_i] = ay[agg_i:a_end,:config.ado].sum(axis=0) / num_elements
                # When there is no size, a zero value is assigned.
                if (s == 0):
                    agg_out[:,f_end-1] = 0.0
                # Transition the start for the next element.
                a_start = a_end
                f_start = f_end
        # Without partial aggregation, we only compute the mean all outputs in each group.
        else:
            a = 0
            for i,s in enumerate(sizes):
                if (s > 0):
                    agg_out[:,i] = ay[a:a+s,:config.ado].sum(axis=0) / s
                else:
                    agg_out[:,i] = 0
                a += s
    # Evaluate the fixed model.
    if (config.mdo > 0):
        if (config.normalize and (config.mdn > 0)):
            # Add the shift term (across all point column vectors).
            x[:config.mdn,:] = (x[:config.mdn,:].T + m.x_shift).T
            # Replace all NaN or Inf values
            x[:config.mdn,:] = np.where(
                np.logical_or(
                    np.isnan(x[:config.mdn,:].astype("float32")),
                    np.isinf(x[:config.mdn,:].astype("float32")),
                ),
                0.0,
                x[:config.mdn,:]
            )
            # Apply the input multiplier.
            if (config.needs_scaling):
                x[:config.mdn,:] = m.x_rescale.T @ x[:config.mdn,:]
        # Evaluate the MLP.
        values = x
        state_values["m_states"] = []
        if (config.mns > 0):
            # Apply input transformation.
            values = np.clip(
                ((values.T @ m.m_input_vecs) + m.m_input_shift).T,
                config.discontinuity, float('inf')
            )
            state_values["m_states"].append(values)
            # Apply internal transformations.
            for i in range(config.mns-1):
                values = np.clip(
                    ((values.T @ m.m_state_vecs[:,:,i]) + m.m_state_shift[:,i]).T,
                    config.discontinuity, float('inf')
                )
                state_values["m_states"].append(values)
        state_values["m_states"] = np.asarray(state_values["m_states"]).T
        # Apply output transformation.
        y = (values.T @ m.m_output_vecs[:,:]).T
    # Apply final normalization.
    if (config.normalize):
        if (config.needs_scaling):
            y[:config.do-config.doe,:] = m.y_rescale.T @ y[:config.do-config.doe,:]
        y[:config.do-config.doe,:] = (y[:config.do-config.doe,:].T + m.y_shift).T
    state_values["y"] = y
    # Return the final values.
    return state_values


# This defines a C structure that can be used to hold this defined type.
class MODEL_CONFIG(ctypes.Structure):
    # (name, ctype) fields for this structure.
    _fields_ = [("adn", ctypes.c_int), ("ade", ctypes.c_int), ("ane", ctypes.c_int), ("ads", ctypes.c_int), ("ans", ctypes.c_int), ("ado", ctypes.c_int), ("adi", ctypes.c_int), ("adso", ctypes.c_int), ("mdn", ctypes.c_int), ("mde", ctypes.c_int), ("mne", ctypes.c_int), ("mds", ctypes.c_int), ("mns", ctypes.c_int), ("mdo", ctypes.c_int), ("mdi", ctypes.c_int), ("mdso", ctypes.c_int), ("do", ctypes.c_int), ("total_size", ctypes.c_long), ("num_vars", ctypes.c_long), ("asev", ctypes.c_long), ("aeev", ctypes.c_long), ("asiv", ctypes.c_long), ("aeiv", ctypes.c_long), ("asis", ctypes.c_long), ("aeis", ctypes.c_long), ("assv", ctypes.c_long), ("aesv", ctypes.c_long), ("asss", ctypes.c_long), ("aess", ctypes.c_long), ("asov", ctypes.c_long), ("aeov", ctypes.c_long), ("msev", ctypes.c_long), ("meev", ctypes.c_long), ("msiv", ctypes.c_long), ("meiv", ctypes.c_long), ("msis", ctypes.c_long), ("meis", ctypes.c_long), ("mssv", ctypes.c_long), ("mesv", ctypes.c_long), ("msss", ctypes.c_long), ("mess", ctypes.c_long), ("msov", ctypes.c_long), ("meov", ctypes.c_long), ("aiss", ctypes.c_long), ("aise", ctypes.c_long), ("aoss", ctypes.c_long), ("aose", ctypes.c_long), ("aims", ctypes.c_long), ("aime", ctypes.c_long), ("aecs", ctypes.c_long), ("aece", ctypes.c_long), ("miss", ctypes.c_long), ("mise", ctypes.c_long), ("moss", ctypes.c_long), ("mose", ctypes.c_long), ("mims", ctypes.c_long), ("mime", ctypes.c_long), ("moms", ctypes.c_long), ("mome", ctypes.c_long), ("mecs", ctypes.c_long), ("mece", ctypes.c_long), ("discontinuity", ctypes.c_float), ("max_step_factor", ctypes.c_float), ("step_factor", ctypes.c_float), ("min_step_factor", ctypes.c_float), ("max_step_component", ctypes.c_float), ("max_curv_component", ctypes.c_float), ("min_curv_component", ctypes.c_float), ("faster_rate", ctypes.c_float), ("slower_rate", ctypes.c_float), ("min_update_ratio", ctypes.c_float), ("update_ratio_step", ctypes.c_float), ("step_mean_change", ctypes.c_float), ("step_curv_change", ctypes.c_float), ("step_ay_change", ctypes.c_float), ("step_emb_change", ctypes.c_float), ("initial_curv_estimate", ctypes.c_float), ("mse_upper_limit", ctypes.c_float), ("min_steps_to_stability", ctypes.c_long), ("max_batch", ctypes.c_long), ("num_threads", ctypes.c_long), ("print_delay_sec", ctypes.c_long), ("steps_taken", ctypes.c_long), ("condition_frequency", ctypes.c_long), ("log_grad_norm_frequency", ctypes.c_long), ("rank_check_frequency", ctypes.c_long), ("num_to_update", ctypes.c_long), ("keep_best", ctypes.c_bool), ("early_stop", ctypes.c_bool), ("basis_replacement", ctypes.c_bool), ("reshuffle", ctypes.c_bool), ("granular_parallelism", ctypes.c_bool), ("partial_aggregation", ctypes.c_bool), ("pairwise_aggregation", ctypes.c_bool), ("ax_normalized", ctypes.c_bool), ("rescale_ax", ctypes.c_bool), ("axi_normalized", ctypes.c_bool), ("ay_normalized", ctypes.c_bool), ("x_normalized", ctypes.c_bool), ("rescale_x", ctypes.c_bool), ("xi_normalized", ctypes.c_bool), ("y_normalized", ctypes.c_bool), ("rescale_y", ctypes.c_bool), ("encode_scaling", ctypes.c_bool), ("normalize", ctypes.c_bool), ("needs_shifting", ctypes.c_bool), ("needs_cleaning", ctypes.c_bool), ("needs_scaling", ctypes.c_bool), ("rwork_size", ctypes.c_long), ("iwork_size", ctypes.c_long), ("lwork_size", ctypes.c_long), ("na", ctypes.c_long), ("nat", ctypes.c_long), ("nm", ctypes.c_long), ("nms", ctypes.c_long), ("nmt", ctypes.c_long), ("i_next", ctypes.c_long), ("i_step", ctypes.c_long), ("i_mult", ctypes.c_long), ("i_mod", ctypes.c_long), ("i_iter", ctypes.c_long), ("smg", ctypes.c_long), ("emg", ctypes.c_long), ("smgm", ctypes.c_long), ("emgm", ctypes.c_long), ("smgc", ctypes.c_long), ("emgc", ctypes.c_long), ("sbm", ctypes.c_long), ("ebm", ctypes.c_long), ("saxb", ctypes.c_long), ("eaxb", ctypes.c_long), ("say", ctypes.c_long), ("eay", ctypes.c_long), ("smxb", ctypes.c_long), ("emxb", ctypes.c_long), ("smyb", ctypes.c_long), ("emyb", ctypes.c_long), ("saet", ctypes.c_long), ("eaet", ctypes.c_long), ("saxs", ctypes.c_long), ("eaxs", ctypes.c_long), ("saxg", ctypes.c_long), ("eaxg", ctypes.c_long), ("sayg", ctypes.c_long), ("eayg", ctypes.c_long), ("smet", ctypes.c_long), ("emet", ctypes.c_long), ("smxs", ctypes.c_long), ("emxs", ctypes.c_long), ("smxg", ctypes.c_long), ("emxg", ctypes.c_long), ("syg", ctypes.c_long), ("eyg", ctypes.c_long), ("saxis", ctypes.c_long), ("eaxis", ctypes.c_long), ("saxir", ctypes.c_long), ("eaxir", ctypes.c_long), ("smxis", ctypes.c_long), ("emxis", ctypes.c_long), ("smxir", ctypes.c_long), ("emxir", ctypes.c_long), ("sal", ctypes.c_long), ("eal", ctypes.c_long), ("sml", ctypes.c_long), ("eml", ctypes.c_long), ("sast", ctypes.c_long), ("east", ctypes.c_long), ("smst", ctypes.c_long), ("emst", ctypes.c_long), ("saxi", ctypes.c_long), ("eaxi", ctypes.c_long), ("smxi", ctypes.c_long), ("emxi", ctypes.c_long), ("ssb", ctypes.c_long), ("esb", ctypes.c_long), ("sao", ctypes.c_long), ("eao", ctypes.c_long), ("smo", ctypes.c_long), ("emo", ctypes.c_long), ("sui", ctypes.c_long), ("eui", ctypes.c_long), ("wint", ctypes.c_long), ("cint", ctypes.c_long), ("wfit", ctypes.c_long), ("cfit", ctypes.c_long), ("wnrm", ctypes.c_long), ("cnrm", ctypes.c_long), ("wgen", ctypes.c_long), ("cgen", ctypes.c_long), ("wemb", ctypes.c_long), ("cemb", ctypes.c_long), ("wevl", ctypes.c_long), ("cevl", ctypes.c_long), ("wgrd", ctypes.c_long), ("cgrd", ctypes.c_long), ("wrat", ctypes.c_long), ("crat", ctypes.c_long), ("wopt", ctypes.c_long), ("copt", ctypes.c_long), ("wcon", ctypes.c_long), ("ccon", ctypes.c_long), ("wrec", ctypes.c_long), ("crec", ctypes.c_long), ("wenc", ctypes.c_long), ("cenc", ctypes.c_long)]
    # Define an "__init__" that can take a class or keyword arguments as input.
    def __init__(self, value=0, **kwargs):
        # From whatever object (or dictionary) was given, assign internal values.
        self.adn = kwargs.get("adn", getattr(value, "adn", value))
        self.ade = kwargs.get("ade", getattr(value, "ade", value))
        self.ane = kwargs.get("ane", getattr(value, "ane", value))
        self.ads = kwargs.get("ads", getattr(value, "ads", value))
        self.ans = kwargs.get("ans", getattr(value, "ans", value))
        self.ado = kwargs.get("ado", getattr(value, "ado", value))
        self.adi = kwargs.get("adi", getattr(value, "adi", value))
        self.adso = kwargs.get("adso", getattr(value, "adso", value))
        self.mdn = kwargs.get("mdn", getattr(value, "mdn", value))
        self.mde = kwargs.get("mde", getattr(value, "mde", value))
        self.mne = kwargs.get("mne", getattr(value, "mne", value))
        self.mds = kwargs.get("mds", getattr(value, "mds", value))
        self.mns = kwargs.get("mns", getattr(value, "mns", value))
        self.mdo = kwargs.get("mdo", getattr(value, "mdo", value))
        self.mdi = kwargs.get("mdi", getattr(value, "mdi", value))
        self.mdso = kwargs.get("mdso", getattr(value, "mdso", value))
        self.do = kwargs.get("do", getattr(value, "do", value))
        self.total_size = kwargs.get("total_size", getattr(value, "total_size", value))
        self.num_vars = kwargs.get("num_vars", getattr(value, "num_vars", value))
        self.asev = kwargs.get("asev", getattr(value, "asev", value))
        self.aeev = kwargs.get("aeev", getattr(value, "aeev", value))
        self.asiv = kwargs.get("asiv", getattr(value, "asiv", value))
        self.aeiv = kwargs.get("aeiv", getattr(value, "aeiv", value))
        self.asis = kwargs.get("asis", getattr(value, "asis", value))
        self.aeis = kwargs.get("aeis", getattr(value, "aeis", value))
        self.assv = kwargs.get("assv", getattr(value, "assv", value))
        self.aesv = kwargs.get("aesv", getattr(value, "aesv", value))
        self.asss = kwargs.get("asss", getattr(value, "asss", value))
        self.aess = kwargs.get("aess", getattr(value, "aess", value))
        self.asov = kwargs.get("asov", getattr(value, "asov", value))
        self.aeov = kwargs.get("aeov", getattr(value, "aeov", value))
        self.msev = kwargs.get("msev", getattr(value, "msev", value))
        self.meev = kwargs.get("meev", getattr(value, "meev", value))
        self.msiv = kwargs.get("msiv", getattr(value, "msiv", value))
        self.meiv = kwargs.get("meiv", getattr(value, "meiv", value))
        self.msis = kwargs.get("msis", getattr(value, "msis", value))
        self.meis = kwargs.get("meis", getattr(value, "meis", value))
        self.mssv = kwargs.get("mssv", getattr(value, "mssv", value))
        self.mesv = kwargs.get("mesv", getattr(value, "mesv", value))
        self.msss = kwargs.get("msss", getattr(value, "msss", value))
        self.mess = kwargs.get("mess", getattr(value, "mess", value))
        self.msov = kwargs.get("msov", getattr(value, "msov", value))
        self.meov = kwargs.get("meov", getattr(value, "meov", value))
        self.aiss = kwargs.get("aiss", getattr(value, "aiss", value))
        self.aise = kwargs.get("aise", getattr(value, "aise", value))
        self.aoss = kwargs.get("aoss", getattr(value, "aoss", value))
        self.aose = kwargs.get("aose", getattr(value, "aose", value))
        self.aims = kwargs.get("aims", getattr(value, "aims", value))
        self.aime = kwargs.get("aime", getattr(value, "aime", value))
        self.aecs = kwargs.get("aecs", getattr(value, "aecs", value))
        self.aece = kwargs.get("aece", getattr(value, "aece", value))
        self.miss = kwargs.get("miss", getattr(value, "miss", value))
        self.mise = kwargs.get("mise", getattr(value, "mise", value))
        self.moss = kwargs.get("moss", getattr(value, "moss", value))
        self.mose = kwargs.get("mose", getattr(value, "mose", value))
        self.mims = kwargs.get("mims", getattr(value, "mims", value))
        self.mime = kwargs.get("mime", getattr(value, "mime", value))
        self.moms = kwargs.get("moms", getattr(value, "moms", value))
        self.mome = kwargs.get("mome", getattr(value, "mome", value))
        self.mecs = kwargs.get("mecs", getattr(value, "mecs", value))
        self.mece = kwargs.get("mece", getattr(value, "mece", value))
        self.discontinuity = kwargs.get("discontinuity", getattr(value, "discontinuity", value))
        self.max_step_factor = kwargs.get("max_step_factor", getattr(value, "max_step_factor", value))
        self.step_factor = kwargs.get("step_factor", getattr(value, "step_factor", value))
        self.min_step_factor = kwargs.get("min_step_factor", getattr(value, "min_step_factor", value))
        self.max_step_component = kwargs.get("max_step_component", getattr(value, "max_step_component", value))
        self.max_curv_component = kwargs.get("max_curv_component", getattr(value, "max_curv_component", value))
        self.min_curv_component = kwargs.get("min_curv_component", getattr(value, "min_curv_component", value))
        self.faster_rate = kwargs.get("faster_rate", getattr(value, "faster_rate", value))
        self.slower_rate = kwargs.get("slower_rate", getattr(value, "slower_rate", value))
        self.min_update_ratio = kwargs.get("min_update_ratio", getattr(value, "min_update_ratio", value))
        self.update_ratio_step = kwargs.get("update_ratio_step", getattr(value, "update_ratio_step", value))
        self.step_mean_change = kwargs.get("step_mean_change", getattr(value, "step_mean_change", value))
        self.step_curv_change = kwargs.get("step_curv_change", getattr(value, "step_curv_change", value))
        self.step_ay_change = kwargs.get("step_ay_change", getattr(value, "step_ay_change", value))
        self.step_emb_change = kwargs.get("step_emb_change", getattr(value, "step_emb_change", value))
        self.initial_curv_estimate = kwargs.get("initial_curv_estimate", getattr(value, "initial_curv_estimate", value))
        self.mse_upper_limit = kwargs.get("mse_upper_limit", getattr(value, "mse_upper_limit", value))
        self.min_steps_to_stability = kwargs.get("min_steps_to_stability", getattr(value, "min_steps_to_stability", value))
        self.max_batch = kwargs.get("max_batch", getattr(value, "max_batch", value))
        self.num_threads = kwargs.get("num_threads", getattr(value, "num_threads", value))
        self.print_delay_sec = kwargs.get("print_delay_sec", getattr(value, "print_delay_sec", value))
        self.steps_taken = kwargs.get("steps_taken", getattr(value, "steps_taken", value))
        self.condition_frequency = kwargs.get("condition_frequency", getattr(value, "condition_frequency", value))
        self.log_grad_norm_frequency = kwargs.get("log_grad_norm_frequency", getattr(value, "log_grad_norm_frequency", value))
        self.rank_check_frequency = kwargs.get("rank_check_frequency", getattr(value, "rank_check_frequency", value))
        self.num_to_update = kwargs.get("num_to_update", getattr(value, "num_to_update", value))
        self.keep_best = kwargs.get("keep_best", getattr(value, "keep_best", value))
        self.early_stop = kwargs.get("early_stop", getattr(value, "early_stop", value))
        self.basis_replacement = kwargs.get("basis_replacement", getattr(value, "basis_replacement", value))
        self.reshuffle = kwargs.get("reshuffle", getattr(value, "reshuffle", value))
        self.granular_parallelism = kwargs.get("granular_parallelism", getattr(value, "granular_parallelism", value))
        self.partial_aggregation = kwargs.get("partial_aggregation", getattr(value, "partial_aggregation", value))
        self.pairwise_aggregation = kwargs.get("pairwise_aggregation", getattr(value, "pairwise_aggregation", value))
        self.ax_normalized = kwargs.get("ax_normalized", getattr(value, "ax_normalized", value))
        self.rescale_ax = kwargs.get("rescale_ax", getattr(value, "rescale_ax", value))
        self.axi_normalized = kwargs.get("axi_normalized", getattr(value, "axi_normalized", value))
        self.ay_normalized = kwargs.get("ay_normalized", getattr(value, "ay_normalized", value))
        self.x_normalized = kwargs.get("x_normalized", getattr(value, "x_normalized", value))
        self.rescale_x = kwargs.get("rescale_x", getattr(value, "rescale_x", value))
        self.xi_normalized = kwargs.get("xi_normalized", getattr(value, "xi_normalized", value))
        self.y_normalized = kwargs.get("y_normalized", getattr(value, "y_normalized", value))
        self.rescale_y = kwargs.get("rescale_y", getattr(value, "rescale_y", value))
        self.encode_scaling = kwargs.get("encode_scaling", getattr(value, "encode_scaling", value))
        self.normalize = kwargs.get("normalize", getattr(value, "normalize", value))
        self.needs_shifting = kwargs.get("needs_shifting", getattr(value, "needs_shifting", value))
        self.needs_cleaning = kwargs.get("needs_cleaning", getattr(value, "needs_cleaning", value))
        self.needs_scaling = kwargs.get("needs_scaling", getattr(value, "needs_scaling", value))
        self.rwork_size = kwargs.get("rwork_size", getattr(value, "rwork_size", value))
        self.iwork_size = kwargs.get("iwork_size", getattr(value, "iwork_size", value))
        self.lwork_size = kwargs.get("lwork_size", getattr(value, "lwork_size", value))
        self.na = kwargs.get("na", getattr(value, "na", value))
        self.nat = kwargs.get("nat", getattr(value, "nat", value))
        self.nm = kwargs.get("nm", getattr(value, "nm", value))
        self.nms = kwargs.get("nms", getattr(value, "nms", value))
        self.nmt = kwargs.get("nmt", getattr(value, "nmt", value))
        self.i_next = kwargs.get("i_next", getattr(value, "i_next", value))
        self.i_step = kwargs.get("i_step", getattr(value, "i_step", value))
        self.i_mult = kwargs.get("i_mult", getattr(value, "i_mult", value))
        self.i_mod = kwargs.get("i_mod", getattr(value, "i_mod", value))
        self.i_iter = kwargs.get("i_iter", getattr(value, "i_iter", value))
        self.smg = kwargs.get("smg", getattr(value, "smg", value))
        self.emg = kwargs.get("emg", getattr(value, "emg", value))
        self.smgm = kwargs.get("smgm", getattr(value, "smgm", value))
        self.emgm = kwargs.get("emgm", getattr(value, "emgm", value))
        self.smgc = kwargs.get("smgc", getattr(value, "smgc", value))
        self.emgc = kwargs.get("emgc", getattr(value, "emgc", value))
        self.sbm = kwargs.get("sbm", getattr(value, "sbm", value))
        self.ebm = kwargs.get("ebm", getattr(value, "ebm", value))
        self.saxb = kwargs.get("saxb", getattr(value, "saxb", value))
        self.eaxb = kwargs.get("eaxb", getattr(value, "eaxb", value))
        self.say = kwargs.get("say", getattr(value, "say", value))
        self.eay = kwargs.get("eay", getattr(value, "eay", value))
        self.smxb = kwargs.get("smxb", getattr(value, "smxb", value))
        self.emxb = kwargs.get("emxb", getattr(value, "emxb", value))
        self.smyb = kwargs.get("smyb", getattr(value, "smyb", value))
        self.emyb = kwargs.get("emyb", getattr(value, "emyb", value))
        self.saet = kwargs.get("saet", getattr(value, "saet", value))
        self.eaet = kwargs.get("eaet", getattr(value, "eaet", value))
        self.saxs = kwargs.get("saxs", getattr(value, "saxs", value))
        self.eaxs = kwargs.get("eaxs", getattr(value, "eaxs", value))
        self.saxg = kwargs.get("saxg", getattr(value, "saxg", value))
        self.eaxg = kwargs.get("eaxg", getattr(value, "eaxg", value))
        self.sayg = kwargs.get("sayg", getattr(value, "sayg", value))
        self.eayg = kwargs.get("eayg", getattr(value, "eayg", value))
        self.smet = kwargs.get("smet", getattr(value, "smet", value))
        self.emet = kwargs.get("emet", getattr(value, "emet", value))
        self.smxs = kwargs.get("smxs", getattr(value, "smxs", value))
        self.emxs = kwargs.get("emxs", getattr(value, "emxs", value))
        self.smxg = kwargs.get("smxg", getattr(value, "smxg", value))
        self.emxg = kwargs.get("emxg", getattr(value, "emxg", value))
        self.syg = kwargs.get("syg", getattr(value, "syg", value))
        self.eyg = kwargs.get("eyg", getattr(value, "eyg", value))
        self.saxis = kwargs.get("saxis", getattr(value, "saxis", value))
        self.eaxis = kwargs.get("eaxis", getattr(value, "eaxis", value))
        self.saxir = kwargs.get("saxir", getattr(value, "saxir", value))
        self.eaxir = kwargs.get("eaxir", getattr(value, "eaxir", value))
        self.smxis = kwargs.get("smxis", getattr(value, "smxis", value))
        self.emxis = kwargs.get("emxis", getattr(value, "emxis", value))
        self.smxir = kwargs.get("smxir", getattr(value, "smxir", value))
        self.emxir = kwargs.get("emxir", getattr(value, "emxir", value))
        self.sal = kwargs.get("sal", getattr(value, "sal", value))
        self.eal = kwargs.get("eal", getattr(value, "eal", value))
        self.sml = kwargs.get("sml", getattr(value, "sml", value))
        self.eml = kwargs.get("eml", getattr(value, "eml", value))
        self.sast = kwargs.get("sast", getattr(value, "sast", value))
        self.east = kwargs.get("east", getattr(value, "east", value))
        self.smst = kwargs.get("smst", getattr(value, "smst", value))
        self.emst = kwargs.get("emst", getattr(value, "emst", value))
        self.saxi = kwargs.get("saxi", getattr(value, "saxi", value))
        self.eaxi = kwargs.get("eaxi", getattr(value, "eaxi", value))
        self.smxi = kwargs.get("smxi", getattr(value, "smxi", value))
        self.emxi = kwargs.get("emxi", getattr(value, "emxi", value))
        self.ssb = kwargs.get("ssb", getattr(value, "ssb", value))
        self.esb = kwargs.get("esb", getattr(value, "esb", value))
        self.sao = kwargs.get("sao", getattr(value, "sao", value))
        self.eao = kwargs.get("eao", getattr(value, "eao", value))
        self.smo = kwargs.get("smo", getattr(value, "smo", value))
        self.emo = kwargs.get("emo", getattr(value, "emo", value))
        self.sui = kwargs.get("sui", getattr(value, "sui", value))
        self.eui = kwargs.get("eui", getattr(value, "eui", value))
        self.wint = kwargs.get("wint", getattr(value, "wint", value))
        self.cint = kwargs.get("cint", getattr(value, "cint", value))
        self.wfit = kwargs.get("wfit", getattr(value, "wfit", value))
        self.cfit = kwargs.get("cfit", getattr(value, "cfit", value))
        self.wnrm = kwargs.get("wnrm", getattr(value, "wnrm", value))
        self.cnrm = kwargs.get("cnrm", getattr(value, "cnrm", value))
        self.wgen = kwargs.get("wgen", getattr(value, "wgen", value))
        self.cgen = kwargs.get("cgen", getattr(value, "cgen", value))
        self.wemb = kwargs.get("wemb", getattr(value, "wemb", value))
        self.cemb = kwargs.get("cemb", getattr(value, "cemb", value))
        self.wevl = kwargs.get("wevl", getattr(value, "wevl", value))
        self.cevl = kwargs.get("cevl", getattr(value, "cevl", value))
        self.wgrd = kwargs.get("wgrd", getattr(value, "wgrd", value))
        self.cgrd = kwargs.get("cgrd", getattr(value, "cgrd", value))
        self.wrat = kwargs.get("wrat", getattr(value, "wrat", value))
        self.crat = kwargs.get("crat", getattr(value, "crat", value))
        self.wopt = kwargs.get("wopt", getattr(value, "wopt", value))
        self.copt = kwargs.get("copt", getattr(value, "copt", value))
        self.wcon = kwargs.get("wcon", getattr(value, "wcon", value))
        self.ccon = kwargs.get("ccon", getattr(value, "ccon", value))
        self.wrec = kwargs.get("wrec", getattr(value, "wrec", value))
        self.crec = kwargs.get("crec", getattr(value, "crec", value))
        self.wenc = kwargs.get("wenc", getattr(value, "wenc", value))
        self.cenc = kwargs.get("cenc", getattr(value, "cenc", value))
    # Define a "__str__" that produces a legible summary of this type.
    def __str__(self):
        s = []
        for (n, t) in self._fields_:
            s.append( n + "=" + str(getattr(self,n)) )
        return "MODEL_CONFIG[" + ", ".join(s) + "]"
    # Define an "__eq__" method that checks equality of all fields.
    def __eq__(self, other):
        for (n, t) in self._fields_:
            if (getattr(self, n) != getattr(other, n, None)):
                return False
        return True



try:
    from .delfort import Delaunay
    from .qhull import Delaunay as QHullDelaunay
except ImportError:
    from tlux.approximate.delaunay.delfort import Delaunay
    from tlux.approximate.delaunay.qhull import Delaunay as QHullDelaunay


# Define a simple test to make sure it is working.
if __name__ == "__main__":
    import numpy as np
    from tlux.plot import Plot
    from tlux.random import well_spaced_box, well_spaced_ball

    # Generate some data.
    x = well_spaced_box(40, 2)
    y = np.cos(10*np.linalg.norm(x, axis=1))

    # Fit a nearest neighbor model.
    m = Delaunay()
    # m = QHullDelaunay()
    m.fit(x, y)

    # Create a visual.
    p = Plot()
    p.add("data", *x.T, y)
    p.add_func("fit", m, *([[0,1]]*x.shape[1]), vectorized=True, plot_points=5000)
    p.show(z_range=[-2,4])



import sys

# Function for printing the help message.
def print_help_message():
    print("""
tlux -- A package aligned with the research compendium of T.C.H.Lux containing a all necessary code to reproduce the work.

### Python

    > import tlux
    > help(tlux)

  Descriptions of the contents follow.

### Command line

    $ python -m tlux [--clean] [--build] [-h] [--help]

  Run the 'tlux' package. When '--clean' is provided, all built directories and compiled
  codes are moved into a stated temporary directory. When '--build' is provided, all
  internal compiled modules are build (using 'fmodpy'). When '-h' or '--help' is provided,
  this message is printed.
    """)

# Clean the build.
if ("--clean" in sys.argv):
    sys.argv.remove("--clean")
    from tlux.setup import clean_all
    clean_all()

# Clean the build.
if ("--build" in sys.argv):
    sys.argv.remove("--build")
    from tlux.setup import build_all
    build_all()

# Provide help.
if (len(sys.argv) > 1) or ("-h" in sys.argv) or ("--help" in sys.argv):
    print_help_message()



# This file defines possible model scenarios that need to be considered
#  when testing. It provides functions for generating unique scenarios
#  with model, data, and work allocations, as well as an iterator that
#  exhaustively iterates over all possible data and model scenarios.

from tlux.approximate.axy.summary import Details
from tlux.approximate.axy.axy import axy as AXY
from tlux.approximate.axy.axy_random import random


# Most combinations of these define the possible uses of the model.
SCENARIO = dict(
    steps = 50,
    ade = None,
    ado = None,
    mde = None,
    aggregator_only = False,
    partial_aggregation = False,
    pairwise_aggregation = False,
    batch_aggregate_constrained = False,
    batch_fixed_constrained = True,
    input_aggregate_categorical = True,
    input_aggregate_numeric = False,
    input_fixed_categorical = False,
    input_fixed_numeric = False,
    model_aggregate_layered = False,
    model_fixed_layered = False,
    output_categorical = False,
    output_numeric = True,
    small_data = False,
    small_model = True,
    threaded = True,
    weighted_output = False,
    weights_dimensioned = False,
)


# Check the exit code values.
def check_code(exit_code, method):
    import os, re
    if (exit_code != 0):
        # Read the reason from the Fortran source file.
        source = ""
        source_file = "../axy.f90"
        if (os.path.exists(source_file)):
            with open(source_file, "r") as f:
                source = f.read()
        reason = re.search(r"INFO\s*=\s*" + str(exit_code) + r"\s*[!][^\n]*\n", source)
        if (reason is not None):
            reason = reason.group().strip()
            reason = reason[reason.index("!")+1:].strip()
        else:
            reason = ""
        # Raise an assertion error.
        assert (exit_code == 0), f"AXY.{method} returned nonzero exit code {exit_code}. {reason}"


# Spawn a new model that is ready to be evaluated.
def spawn_model(adn, mdn, mdo, ade=0, ane=0, ads=3, ans=2, ado=None, mde=0, mne=0, mds=3, mns=2,
                doe=None, noe=0, seed=0, num_threads=30, initial_shift_range=1.0, initial_output_scale=0.1):
    import numpy as np
    # Create a new config.
    config = AXY.new_model_config(
        adn = adn,
        ade = ade,
        ane = ane,
        ads = ads,
        ans = ans,
        ado = ado,
        mdn = mdn,
        mde = mde,
        mne = mne,
        mds = mds,
        mns = mns,
        mdo = mdo,
        noe = noe,
        doe = doe,
        num_threads = num_threads,
    )
    # Initialize the model.
    model = np.zeros(config.total_size, dtype="float32")
    AXY.init_model(config, model, seed=seed,
                   initial_shift_range=initial_shift_range,
                   initial_output_scale=initial_output_scale)
    # By default, make sure reshuffling is disabled (for determinism with aggregate inputs).
    config.reshuffle = False
    # Return the config and the model.
    return config, model


# Given a scenario, generate a model and data to match the scenario.
def gen_config_data(scenario=None, seed=0, default=SCENARIO, **scenario_kwargs):
    # Update the default scenario with the provided one.
    settings = default.copy()
    if (scenario is not None):
        settings.update(scenario)
    scenario = settings
    scenario.update(scenario_kwargs)
    # Problem size.
    if scenario["small_data"]:
        nm_in = scenario.get('nm_in', 10)
        na_in = scenario.get('na_in', 100)
        na_range = scenario.get('na_range', (1, 5))
        nm_range = scenario.get('nm_range', (1, 5))
    else:
        nm_in = scenario.get('nm_in', 10000)
        na_in = scenario.get('na_in', 1000000)
        na_range = scenario.get('na_range', (1, 5000))
        nm_range = scenario.get('nm_range', (1, 5000))
    # Fixed batch constrained (cannot fit all fixed data in at once).
    if scenario["batch_fixed_constrained"]:
        nm = scenario.get('nm', nm_in // 5)
    else:
        nm = scenario.get('nm', nm_in)
    # Aggregate batch constrained (cannot fit all aggregate data in at once).
    if scenario["batch_aggregate_constrained"]:
        na = scenario.get('na', na_in // 5)
    else:
        na = scenario.get('na', na_in)
    # Aggregate numeric input.
    if scenario["input_aggregate_numeric"]:
        if scenario["small_data"]:
            adn = scenario.get('adn', 2)
        else:
            adn = scenario.get('adn', 20)
    else:
        scenario.pop("adn", None)
        adn = 0 # scenario.get('adn', 0)
    # Aggregate categorical input.
    if scenario["input_aggregate_categorical"]:
        if scenario["small_data"]:
            ane = scenario.get('ane', 5)
            adi = scenario.get('adi', 1)
        else:
            ane = scenario.get('ane', 5000)
            adi = scenario.get('adi', 5)
    else:
        scenario.pop("ane", None)
        scenario.pop("adi", None)
        ane = 0 # scenario.get('ane', 0)
        adi = 0 # scenario.get('adi', 0)
    # Aggregate model layered.
    if scenario["model_aggregate_layered"]:
        if scenario["small_model"]:
            ans = scenario.get('ans', 2)
            ads = scenario.get('ads', 3)
        else:
            ans = scenario.get('ans', 20)
            ads = scenario.get('ads', 64)
    else:
        ans = scenario.get('ans', 0)
        ads = scenario.get('ads', 0)
    # Fixed numeric input.
    if scenario["input_fixed_numeric"]:
        if scenario["small_data"]:
            mdn = scenario.get('mdn', 2)
        else:
            mdn = scenario.get('mdn', 20)
    else:
        mdn = scenario.get('mdn', 0)
    # Fixed categorical input.
    if scenario["input_fixed_categorical"]:
        if scenario["small_data"]:
            mne = scenario.get('mne', 5)
            mdi = scenario.get('mdi', 1)
        else:
            mne = scenario.get('mne', 5000)
            mdi = scenario.get('mdi', 5)
    else:
        scenario.pop("mne", None)
        scenario.pop("mdi", None)
        mne = 0 # scenario.get('mne', 0)
        mdi = 0 # scenario.get('mdi', 0)
    # Fixed model layered.
    if scenario["model_fixed_layered"]:
        if scenario["small_model"]:
            mns = scenario.get('mns', 2)
            mds = scenario.get('mds', 3)
        else:
            mns = scenario.get('mns', 20)
            mds = scenario.get('mds', 64)
    else:
        mns = scenario.get('mns', 0)
        mds = scenario.get('mds', 0)
    # Numeric output.
    if scenario["output_numeric"]:
        if scenario["small_data"]:
            ydn = scenario.get('ydn', 2)
        else:
            ydn = scenario.get('ydn', 20)
    else:
        ydn = scenario.get('ydn', 0)
    # Categorical output.
    if scenario["output_categorical"]:
        if scenario["small_data"]:
            yne = scenario.get('yne', 5)
            ydi = scenario.get('ydi', 1)
        else:
            yne = scenario.get('yne', 5000)
            ydi = scenario.get('ydi', 5)
    else:
        yne = scenario.get('yne', 0)
        ydi = scenario.get('ydi', 0)
    # Aggregator only.
    if scenario["aggregator_only"]:
        mns = scenario.get('mns', 0)
        mds = scenario.get('mds', 0)
        mdo = scenario.get('mdo', -1)
    else:
        mdo = ydn + yne
    # Weighted output dimensioned.
    if scenario["weighted_output"]:
        if scenario["weights_dimensioned"]:
            ywd = scenario.get('ywd', ydn + yne)
        else:
            ywd = scenario.get('ywd', 1)
    else:
        ywd = scenario.get('ywd', 0)
    # Multithreading.
    if scenario["threaded"]:
        num_threads = scenario.get('num_threads', None)
    else:
        num_threads = scenario.get('num_threads', 1)
    # Training steps.
    steps = scenario['steps']
    # Generate the model config.
    config = AXY.new_model_config(
        adn=adn,
        ade=scenario.get("ade",None),
        ane=ane,
        ads=ads,
        ans=ans,
        ado=scenario.get("ado",None),
        mdn=mdn,
        mde=scenario.get("mde",None),
        mne=mne,
        mds=mds,
        mns=mns,
        mdo=mdo,
        noe=yne,
        num_threads=num_threads,
    )
    # Generate a fit configuration.
    config.partial_aggregation = scenario.get('partial_aggregation', False)
    config = AXY.new_fit_config(
        nm=nm,
        na=na,
        nmt=nm_in,
        nat=na_in,
        adi=adi,
        mdi=mdi,
        odi=ydi,
        seed=seed,
        config=config,
    )
    # Set any config values that were not already declared.
    scenario.pop("mdi", None)
    scenario.pop("adi", None)
    scenario.pop("adn", None)
    for (name, value) in scenario.items():
        # These items should not be overwritten.
        if (name in {"num_threads"}): continue
        # Any other items are probably explicitly being declared by the user.
        if (hasattr(config, name) and (value is not None) and (getattr(config, name) != value)):
            # print(f" overwriting '{name}' with {value}")
            setattr(config, name, value)
    # Seed the randomness for repeatability.
    import numpy as np
    np.random.seed(seed)
    # Generate data that matches these specifications.
    ftype = dict(order="F", dtype="float32")
    itype = dict(order="F", dtype="int32")
    ltype = dict(order="F", dtype="int64")
    asarray = lambda a, t: np.asarray(a, **t)
    ax_in = asarray(np.random.normal(size=(adn, config.nat)), ftype)
    axi_in = asarray(np.random.randint(*na_range, size=(adi, config.nat)), ltype)
    sizes_in = asarray(np.random.randint(0, max(1,round(2*(config.nat / nm_in))), size=(nm_in if config.nat > 0 else 0,)), ltype)
    x_in = asarray(np.random.normal(size=(mdn, nm_in)), ftype)
    xi_in = asarray(np.random.randint(*nm_range, size=(mdi, nm_in)), ltype)
    y_in = asarray(np.random.normal(size=(config.do-config.doe, nm_in)), ftype)
    if (config.noe > 0): yi_in = asarray(np.random.randint(1, config.noe, size=(ydi, nm_in)), ltype)
    else:                yi_in = np.zeros((0,nm_in), **ltype)
    yw_in = asarray(np.random.normal(size=(ywd, nm_in)), ftype)
    # Set two of the sizes to zero (to make sure there are zeros in there.
    if (len(sizes_in) > 0):
        sizes_in[1*len(sizes_in) // 3] = 0
        sizes_in[2*len(sizes_in) // 3] = 0
    # Adjust the sizes to make sure the sum is the correct value.
    ssum = sum(sizes_in)
    i = 0
    while (ssum != config.nat):
        if (ssum < config.nat):
            ssum += 1
            sizes_in[i] += 1
        elif (ssum > config.nat) and (sizes_in[i] > 0):
            ssum -= 1
            sizes_in[i] -= 1
        i = (i + 1) % len(sizes_in)
    # Adjust the "na" value based on pairwise aggregation.
    pairwise = scenario.get("pairwise_aggregation",False)
    if (pairwise):
        config.na = sum(sizes_in**2)
        if (scenario["batch_aggregate_constrained"]):
            config.na = config.na // 5
        config = AXY.new_fit_config(
            nm=nm,
            na=config.na,
            nmt=nm_in,
            nat=na_in,
            adi=adi,
            mdi=mdi,
            odi=ydi,
            seed=seed,
            config=config,
        )
    # Generate the memory and references to specific data holders.
    details = Details(config, steps, ywd=ywd)
    # Initialize the model.
    AXY.init_model(config, details.model, seed=seed)
    # Data holders.
    ax = details.ax
    axi = details.axi
    sizes = details.sizes
    x = details.x
    xi = details.xi
    y = details.y
    yi = details.yi
    yw = details.yw
    # Return the config and data.
    return (
        config, details,
        dict(ax_in=ax_in, axi_in=axi_in, sizes_in=sizes_in, x_in=x_in, xi_in=xi_in, y_in=y_in, yi_in=yi_in, yw_in=yw_in),
        dict(ax=ax, axi=axi, sizes=sizes, x=x, xi=xi, y=y, yi=yi, yw=yw),
    )


# Define a scenario generator.
def scenario_generator(scenario=SCENARIO, randomized=True, seed=None):
    # If this is randomized, overwrite the "range" function with a random one.
    if randomized:
        if (seed is not None):
            import random
            random.seed(0)
        from tlux.random import random_range
        range = random_range
    else:
        import builtins
        range = builtins.range
    # Get all keys.
    keys = sorted((k for k,v in scenario.items() if (type(v) is bool)))
    # Iterate over all binary pairs of keys.
    for i in range(2**len(keys)):
        # Start with the default scenario.
        s = scenario.copy()
        # Decide which combination of keys to set to True (using binary digits)
        bin_str = bin(i)[2:][::-1] + '0'*len(keys)
        for v,n in zip(bin_str, keys):
            s[n] = (v == '1')
        # Skip any configurations that affect the model when it is absent.
        if (s["aggregator_only"] and (
                s["input_fixed_numeric"] or
                s["input_fixed_categorical"] or
                s["model_fixed_layered"] or
                s["batch_fixed_constrained"]
        )):
            pass
        # Skip the invalid configuration where outputs are NOT weighted and
        #  "weights_dimensioned" is True, that is meaningless since there are no weights.
        elif ((not s["weighted_output"]) and (s["weights_dimensioned"])):
            pass
        # Skip the invalid configuration where there are no outputs.
        elif (not (s["output_numeric"] or s["output_categorical"])):
            pass
        # Skip the invalid configuration for "large model" and NOT "model layered"
        elif not (s["small_model"] or 
                  s["model_fixed_layered"] or
                  s["model_aggregate_layered"]):
            pass
        # Skip scenarios involving the aggregator when it is not present.
        elif ((not s["input_aggregate_numeric"]) and
              (not s["input_aggregate_categorical"]) and
              (s["batch_aggregate_constrained"] or s["model_aggregate_layered"] or s["pairwise_aggregation"] or s["partial_aggregation"])):
            pass
        # Skip when there is no input data at all.
        elif ((not s["input_aggregate_numeric"]) and
              (not s["input_aggregate_categorical"]) and
              (not s["input_fixed_numeric"]) and
              (not s["input_fixed_categorical"])):
            pass
        # Otherwise, this is a valid scenario.
        else:
            yield s


# Initialize an aggregate iterator given a config and list of sizes.
def initialize_agg_iterator(config, agg_iterators, sizes_in, seed=None):
    for i in range(agg_iterators.shape[1]):
        if (sizes_in[i] == 0):
            agg_iterators[:,i] = 0
        else:
            agg_iterators[0,i] = sizes_in[i]
            if config.pairwise_aggregation:
                agg_iterators[0,i] = agg_iterators[0,i]**2
            agg_iterators[1:,i] = random.initialize_iterator(
                i_limit=agg_iterators[0,i], seed=seed
            )




# A memory mapped Fasttext model (slightly slower, but uses much less memory).
class MemmapFasttext:

    # Load a memory mapped fasttext model. Can be provided any of:
    #   path = "<fasttext-model>.bin"  -->  Will load, convert to memmap, save memmap in adjacent path, and load memmap.
    #   path = "<memmap-model-path>"  -->  Will load that memmamp model path directly.
    #   model = fasttext.Fasttext  -->  Will convert to memmap, save in a temporary directory (unless path provided), and load memmap.
    # 
    # Optional:
    #   overwrite  -->  Provide in the case that the memmap directory already exists and should be overwritten.
    #   max_cache_size  -->  Provide an integer or None to set a positive limit to functools.lru_cache around internal methods.
    # 
    def __init__(self, path="", model=None, overwrite=False, max_cache_size=0):
        from tlux.profiling import mem_use
        # If a ".bin" file was provided, then read it and generate a 
        #   MemmapFasttext of that pure Fasttext model in an adjacent path.
        if (path.endswith(".bin")):
            og_path = path
            path += ".mmap"
            import os
            if os.path.exists(path):
                model = None
            else:
                import sys
                if os.path.dirname(__file__) in sys.path:
                    sys.path.remove(os.path.dirname(__file__))
                import fasttext
                model = fasttext.load_model(og_path)
        # This might be an attempt to save a model and convert it to a MemmapFasttext.            
        if ((type(path) is not str) and (type(model) is None)):
            model = path
            path = ""
        if (model is not None):
            if (len(path) == 0):
                import tempfile
                self.temp_dir = tempfile.TemporaryDirectory()
                path = self.temp_dir.name
            self.save_memmap(path=path, model=model, overwrite=overwrite)
        # Store the path locally.
        self.path = path
        # Load the model from the provided path.
        ( 
            config_path, subword_vecs_path, word_vecs_path,
            word_strings_path, word_starts_path, word_ends_path,
            word_order_path
        ) = self.get_paths(path)
        # Read the configuration.
        with open(config_path) as f:
            import json
            for key, value in json.loads(f.read()).items():
                if (key not in {"bucket", "minn", "maxn", "wordsn", "subwordsn", "dim", "real_type", "char_type", "int_type"}):
                    raise(ValueError(f"Unrecognized key in model JSON, '{key}'."))
                setattr(self, key, value)
        # Load the various memory mapped files.
        import numpy as np
        self.subwords_matrix = np.memmap(subword_vecs_path, dtype=self.real_type, shape=(self.subwordsn, self.dim), mode="r")
        self.words_matrix = np.memmap(word_vecs_path, dtype=self.real_type, shape=(self.wordsn, self.dim), mode="r")
        self.word_strings = np.memmap(word_strings_path, dtype=self.char_type, mode="r")
        self.word_start = np.memmap(word_starts_path, dtype=self.int_type, mode="r")
        self.word_end = np.memmap(word_ends_path, dtype=self.int_type, mode="r")
        self.word_order = np.memmap(word_order_path, dtype=self.int_type, mode="r")
        # If caching should be enabled, then decorate internal methods.
        if max_cache_size is None:
            from functools import cache
            self.get_word = cache(self.get_word_at)
            self.get_word_id = cache(self.get_word_id)
            self.get_subword_id = cache(self.get_subword_id)
            self.get_subwords = cache(self.get_subwords)
            self.get_word_vector = cache(self.get_word_vector)
        elif max_cache_size > 0:
            from functools import lru_cache
            self.get_word = lru_cache(maxsize=max_cache_size)(self.get_word_at)
            self.get_word_id = lru_cache(maxsize=max_cache_size)(self.get_word_id)
            self.get_subword_id = lru_cache(maxsize=max_cache_size)(self.get_subword_id)
            self.get_subwords = lru_cache(maxsize=max_cache_size)(self.get_subwords)
            self.get_word_vector = lru_cache(maxsize=max_cache_size)(self.get_word_vector)


    # Return the list of all words (in the original fasttext order).
    @property
    def words(self):
        if not hasattr(self, "_words"):
            self._words = [
                self.get_word_at(i) for i in
                sorted(range(self.wordsn), key=lambda i: self.word_order[i])
            ]
        return self._words


    # Given an index, get the word at that index out of the sorted-order list of words.
    def get_word_at(self, i):
        return "".join(map(chr, self.word_strings[self.word_start[i]:self.word_end[i]]))


    # Get the index associated with a word, return None if there is not an exact match.
    def get_word_id(self, word):
        # Perform a binary search to try and find the word.
        low = 0
        high = self.wordsn
        index = (low + high) // 2
        while high != low:
            word_at_index = self.get_word_at(index)
            if (len(word_at_index) > len(word)):  high = index
            elif (len(word_at_index) < len(word)): low = index + 1
            else:
                if (word_at_index > word):  high = index
                elif (word_at_index < word): low = index + 1
                else: return self.word_order[index]
            index = (low + high) // 2
        # If the word is not found, then it does not exist in the vocabulary.
        return None


    # Mimic the Fasttext hash function as well as the dictionary key lookup
    #  function to produce the index of a subword given its string.
    def get_subword_id(self, string):
        h = 2166136261 # Use Fasttext initial value.
        for c in string:
            h = h ^ ord(c) # Use Fasttext XOR strategy (for bit-level diversity).
            h = h * 16777619 # Use Fasttext multiplier.
            h = h & 0xFFFFFFFF # Keep in uint32 range to mimic C++.
        return h % self.bucket + self.wordsn # Perform the mod and offset used by the Fasttext Dictionary.


    # Given a word, produce the list of subwords and subword IDs contained in the word.
    def get_subwords(self, word):
        subwords = []
        subword_ids = []
        # Add the index of the word if it is known.
        word_id = self.get_word_id(word)
        if word_id is not None:
            subwords.append(word)
            subword_ids.append(word_id)
        # Get all subwords.
        word = "<" + word + ">"
        for i in range(0, len(word)-self.minn+1):
            for j in range(i+self.minn,min(i+self.maxn,len(word))+1):
                subword = word[i:j]
                subwords.append(subword)
                subword_ids.append(self.get_subword_id(subword))
        # Return the list of subwords and IDs.
        return subwords, subword_ids


    # Get the word vector for this word.
    def get_word_vector(self, word):
        import numpy as np
        # First try and retrieve a known word.
        word_id = self.get_word_id(word)
        if (word_id is not None):
            return self.words_matrix[word_id]
        # Otherwise, get the subwords and compute the embedding.
        subwords, subword_ids = self.get_subwords(word)
        vec = np.zeros(self.dim, dtype=float)
        for i in subword_ids:
            vec += self.subwords_matrix[i]
        if (len(subword_ids) > 0):
            vec /= len(subword_ids)
        # Return the vector representation.
        return vec


    # Generate the names of all file paths given the MemmapFasttext directory path.
    @classmethod
    def get_paths(cls, dirname):
        import os
        config_path = os.path.join(dirname, "model_config.json")
        subword_vecs_path = os.path.join(dirname, "subword_vectors.memmap")
        word_vecs_path = os.path.join(dirname, "word_vectors.memmap")
        word_strings_path = os.path.join(dirname, "word_strings.memmap")
        word_starts_path = os.path.join(dirname, "word_start.memmap")
        word_ends_path = os.path.join(dirname, "word_end.memmap")
        word_order_path = os.path.join(dirname, "word_order.memmap")
        return config_path, subword_vecs_path, word_vecs_path, word_strings_path, word_starts_path, word_ends_path, word_order_path


    # Given a Fasttext model, save all necessary data to load it as an mmap.
    @classmethod
    def save_memmap(cls, path, model, overwrite=False, char_type="uint32", int_type="uint32"):
        import os, json
        import numpy as np
        # Make the directory.
        os.makedirs(path, exist_ok=overwrite)
        # Get all file paths.
        dirname = path
        ( 
            config_path, subword_vecs_path, word_vecs_path,
            word_strings_path, word_starts_path, word_ends_path,
            word_order_path
        ) = cls.get_paths(dirname)
        # Save the subwords matrix.
        matrix = model.get_input_matrix()
        real_type = matrix.dtype.name
        subwords_matrix = np.memmap(
            subword_vecs_path, dtype=matrix.dtype,
            mode="w+", shape=matrix.shape
        )
        subwordsn = matrix.shape[0]
        dim = matrix.shape[1]
        subwords_matrix[:,:] = matrix[:,:]
        # We are done with the source matrix.
        del matrix
        # Trigger a write by flushing the memmap array.
        subwords_matrix.flush()
        # Reloade the matrix in "read" mode.
        subwords_matrix = np.memmap(subword_vecs_path, dtype=real_type, mode="r")
        # Save the words matrix.
        wordsn = len(model.words)
        # Vectors.
        words_matrix = np.memmap(
            word_vecs_path, dtype=subwords_matrix.dtype,
            mode="w+", shape=(wordsn, dim)
        )
        all_words_len = 0
        for i,w in enumerate(model.words):
            all_words_len += len(w)
            words_matrix[i,:] = model.get_word_vector(w)
        words_matrix.flush()
        # Strings.
        word_strings = np.memmap(word_strings_path, dtype=char_type, mode="w+", shape=(all_words_len))
        word_start = np.memmap(word_starts_path, dtype=int_type, mode="w+", shape=(wordsn))
        word_end = np.memmap(word_ends_path, dtype=int_type, mode="w+", shape=(wordsn))
        word_order = np.memmap(word_order_path, dtype=int_type, mode="w+", shape=(wordsn))
        # Iterate over words, writing each one.
        start_index = 0
        word_order[:] = sorted(range(wordsn), key=lambda i: (len(model.words[i]), model.words[i]))
        for i,o in enumerate(word_order):
            w = model.words[o]
            end_index = start_index + len(w)
            word_start[i] = start_index
            word_end[i] = end_index
            word_strings[start_index:end_index] = list(map(ord, w))
            start_index = end_index
        word_strings.flush()
        word_start.flush()
        word_end.flush()
        word_order.flush()
        char_type = word_strings.dtype.name
        int_type = word_order.dtype.name
        # Collect and write the details about the model to a JSON.
        args = model.f.getArgs()
        bucket = args.bucket
        minn = args.minn
        maxn = args.maxn
        with open(config_path, "w") as f:
            f.write(json.dumps(dict(
                bucket=bucket, minn=minn, maxn=maxn, wordsn=wordsn, subwordsn=subwordsn,
                dim=dim, real_type=real_type, char_type=char_type, int_type=int_type
            )))



# Build a fasttext model over this file.
def _test_build_model():
    # Disable standard error to prevent fasttext C++ lib from printing.
    import os, sys
    with open(os.devnull, "w") as f:
        os.dup2(f.fileno(), sys.stderr.fileno())
    # Build a fasttext model over this file.
    import fasttext
    model = fasttext.train_unsupervised(__file__, model='skipgram')
    model.save_model("test.bin")


# Compare the memmap model and the fasttext model.
def _compare_mmap_fasttext():
    # Pop this file's directory from the path to avoid name issues with "fasttext" module.
    import os, sys
    if os.path.dirname(__file__) in sys.path:
        sys.path.remove(os.path.dirname(__file__))

    print(" testing memmap fasttext.. building model", end=".. ")
    import multiprocessing
    builder = multiprocessing.Process(target=_test_build_model)
    builder.start()
    builder.join()
    builder.close()

    # Use the profiling function to track memory usage.
    from tlux.profiling import mem_use

    # Get Memmap model memory usage.
    memory_1 = mem_use()
    mmap = MemmapFasttext("test.bin", max_cache_size=None)
    memory_1 = mem_use() - memory_1

    # Get Fasttext model memory usage.
    memory_2 = mem_use()
    import fasttext
    model = fasttext.load_model("test.bin")
    memory_2 = mem_use() - memory_2
    print("size", end=".. ")
    assert (memory_1 / memory_2 < 0.2), f"Memmap model had a larger than expected size ({memory_1}MB) relative to original Fasttext model ({memory_2}MB)."

    # Verify that the words lists are the same.
    print("words", end=".. ")
    assert (tuple(mmap.words) == tuple(model.words)), f"Memmap words did not match fasttext words."

    # Compute the sum total difference between the word vectors.
    total_diff = 0.0
    for w in model.words:
        total_diff += sum(abs(model.get_word_vector(w) - mmap.get_word_vector(w)))
    print("vecs", end=".. ")
    assert (total_diff < 1e-06), f"Total difference between fasttext and memmap was too large.\n  {total_diff}"

    # Compute and compare the embeddings for random words.
    with open(__file__) as f:
        words = f.read().split()

    # Compute difference when embedding strange words.
    max_diff = 0.0
    for w in words:
        max_diff = max(max_diff, max(abs(model.get_word_vector(w) - mmap.get_word_vector(w))))
    print("novel vecs", end=".. ")
    assert (max_diff < 1e-06), f"Maximum difference between fasttext and memmap was too large.\n  {max_diff}"

    # Remove the build models.
    print("cleaning up", end=".. ")
    os.remove("test.bin")
    import shutil
    shutil.rmtree(mmap.path)

    print("passed", end="!\n")

if __name__ == "__main__":
    _compare_mmap_fasttext()



import ctypes
import os
import itertools
import functools
import math
import multiprocessing
import numpy as np


_lib_dir = os.path.dirname(os.path.abspath(__file__))
_lib_path = os.path.join(_lib_dir, "unique.so")

# Import the C library
try:
    clib = ctypes.cdll.LoadLibrary(_lib_path)
except:
    from tlux.setup import build_unique
    clib = build_unique()


# Function for converting arrays into a C compatible type.
# Inputs:
#   array - Either an np.ndarray, or any python iterable with a __len__. For an
#           ndarray of any fixed-width type, will be converted with the builtin
#           'tobytes'. For arrays of objects or generic iterables, all data will
#           be treated as pythone 'str' objects.
# Outputs:
#   array - A np.ndarray that references the same memory as the following ctypes objects.
#   voidp - A ctypes.c_void_p object that points to byte
#           representations of the contents of 'array'
#   length - The ctypes.c_long number of elements in the array 'voidp'.
#   width - The ctypes.c_int bytes per data object referenced in 'voidp' if
#           the data type has fixed width, otherwise the value -1.
#   info - A ctypes.c_char_p that contains the name of the data type stored in 'voidp'.
def to_bytes(array):
    # If this is a numpy array, then flatten it first (to iterate over elements).
    if (isinstance(array, np.ndarray) and (len(array.shape) > 1)):
        array = array.flatten()
    # If this is a numpy array and it is not contiguous, reallocate.
    if (isinstance(array, np.ndarray) and (not array.data.c_contiguous)):
        array = np.asarray(array, order="C")
    # If this is a numpy array, 
    if (isinstance(array, np.ndarray) 
        and (array.dtype.name != 'object')):
        # Convert it directly into the correctly sized C array.
        voidp = ctypes.c_void_p(array.ctypes.data)
        length = ctypes.c_long(len(array))
        width = ctypes.c_int(array.dtype.itemsize)
        info = ctypes.c_char_p(array.dtype.name.encode())
    else:
        # This is a structure of python objects, convert everything to strings, encode
        #  those, put bytes in an array of ctypes.c_char_p, cast that to c_void_p.
        charp = (ctypes.c_char_p * len(array))(*(str(i).encode() for i in array))
        voidp = ctypes.cast(charp, ctypes.c_void_p)
        length = ctypes.c_long(len(array))
        width = ctypes.c_int(-1)
        info = ctypes.c_char_p(b'object')
        try:    array = np.fromiter((s for s in charp), dtype=object)
        except: array = np.asarray([s for s in charp], dtype=object)
    # Return the final values (be sure to keep a reference to the 'array' so it isn't freed).
    return array, voidp, length, width, info


# Function for converting objects created with 'to_bytes' back into the
# source object. Inputs:
#   array - A vestigial input (can be None) in case 'to_bytes' needed a safety reference.
#   voidp - A ctypes.c_void_p object that points to an array of data.
#   length - The ctypes.c_long number of elements in the array 'voidp'.
#   width - The ctypes.c_int bytes per data object referenced in 'voidp' if the
#           data type has fixed width, otherwise the value -1.
#   info - A ctypes.c_char_p that contains the name of the data type stored in 'voidp'.
# 
# Outputs:
#   array - A np.ndarray object containing the appropriately formed data (either the
#           correct fixed-width numbers, or python 'str' objects for strings).
def from_bytes(array, voidp, length, width, info):
    # Appropriately load the data from bytes.
    if (width.value == -1):
        # If the width is -1, the data is of the 'str' type.
        # Extract the bytes and decode them into 'str' objects.
        charp = (ctypes.c_char_p * length.value).from_address(voidp.value)
        # Convert the list of 'str' objects into a numpy array.
        try:    array = np.fromiter((s for s in charp), dtype=object)
        except: array = np.asarray([s for s in charp], dtype=object)
    else:
        # If the width is not -1, the data is of a fixed-width type.
        # Create a numpy array of the appropriate data type from the raw bytes.
        voidp = (ctypes.c_void_p * length.value).from_address(voidp.value)
        array = np.frombuffer(voidp, dtype=info.value.decode(), count=length.value)
    return array


# Holder for array data (that may have been created with custom C library).
#  Includes functionality to make it behave somewhat like a NumPy array.
class ByteArray:
    def __init__(self, array, voidp, length, width, info, needs_freeing, print_limit=20):
        self.array = array
        self.voidp = voidp
        self.length = length
        self.width = width
        self.info = info
        self.needs_freeing = needs_freeing
        self.print_limit = print_limit
    # Cleanup for C-allocated data.
    def __del__(self):
        if (hasattr(self, "needs_freeing") and (self.needs_freeing)):
            clib.free_unique(
                ctypes.byref(self.voidp),
                ctypes.byref(self.length),
                ctypes.byref(self.width)
            )
    # List-like behaviors.
    def __getitem__(self, i):
        if (self.width.value == -1):
            return self.array[i].decode()
        else:
            return self.array[i].item()
    def __len__(self):
        return self.array.size
    def __repr__(self):
        return f"<{type(self).__name__}[{len(self)}] of '{self.info.value.decode()}' at {ctypes.cast(self.voidp, ctypes.c_void_p).value}>"
    def __str__(self):
        s = "["
        # If this is an object with lots of entries, split in half and fill with "...".
        if (len(self) > self.print_limit):
            s += " ".join((
                " ".join((repr(self[i]) for i in range(min(len(self),self.print_limit//2)))),
                "...",
                " ".join((repr(self[i]) for i in range(max(0,len(self)-self.print_limit//2),len(self)))),
            ))
        # Otherwise this object can be shown in full.
        else:
            s += " ".join((repr(self[i]) for i in range(len(self))))
        s += "]"
        return s    
    # Numpy-like behaviors.
    @property
    def size(self):
        return len(self)
    @property
    def shape(self):
        return (len(self),)
    @property
    def dtype(self):
        return type(self)
    # WARNING: This 'tolist' implementation includes metadata, designed only to be consumed by 'fromlist' of this class.
    def tolist(self):
        return [
            [self[i] for i in range(len(self))],
            self.length.value,
            self.width.value,
            self.info.value.decode(),
        ]
    # Method for loading a value created with 'tolist'.
    @staticmethod
    def fromlist(list_value):
        # Unpack expected items from list.
        array, length, width, info = list_value
        # Convert the array into the expected type.
        array = np.asarray(array, dtype=info)
        # Convert the array (of the expected type) into C-compatible bytes.
        array, voidp, length, width, info = to_bytes(array)
        # For types that are not strings, we need to convert 'voidp' into an array of pointers.
        if (width.value != -1):
            voidp = (ctypes.c_void_p * length.value)(*(
                ctypes.c_void_p(voidp.value + i*width.value)
                for i in range(length.value)
            ))
        return ByteArray(array, voidp, length, width, info, needs_freeing=False)


# Return the sorted unique elements in an array.
def unique(array):
    # Create a C compatible version of this array.
    array, voidp, length, width, info = to_bytes(array)
    # return sorted(set(map(str,array)))
    # 
    # Call the parallel sort function.
    num_unique = ctypes.c_long()
    sorted_unique_pointer = ctypes.POINTER(ctypes.POINTER(ctypes.c_char))()
    clib.unique(voidp, length, width, ctypes.byref(num_unique), ctypes.byref(sorted_unique_pointer))
    # Check for errors from the 'unique' routine.
    if (num_unique.value == -1):
        raise(TypeError(f"The 'unique' C library cannot handle items of byte-width {width.value} (type '{info.value.decode}'). Consider converting to python strings or requesting support via a GitHub Issue."))
    # Get the pointer to the raw array of unique values (contiguously packed).
    sorted_unique = ctypes.cast(
        sorted_unique_pointer if (width.value == -1) else sorted_unique_pointer.contents,
        ctypes.c_void_p
    )
    # Transfer the data over to an object that behaves like a list or numpy array,
    # while also including code to free the memory allocated internally when it is deleted.
    unique_array = from_bytes(None, sorted_unique, num_unique, width, info)
    return ByteArray(unique_array, sorted_unique_pointer, num_unique, width, info, needs_freeing=True)


# Given an array and a mapping that is the result of calling "unique",
#  return an array of integers where all elements of "array" hvae been,
#  replaced with either their reference integer in mapping or 0 for unknown.
def to_int(array, mapping):
    array = np.asarray(array, dtype=mapping.array.dtype).flatten()
    num_array = np.zeros(array.shape, dtype="long")
    # Get the C-compatible representation of the data in 'array'.
    _, voidp, length, width, info = to_bytes(array)
    clib.to_int(
        # Configuration.
        mapping.width,
        # Input data.
        length,
        voidp,
        # Mapping.
        mapping.length,
        ctypes.cast(mapping.voidp, ctypes.c_void_p),
        # Output.
        ctypes.c_void_p(num_array.ctypes.data),
    )
    return num_array



# Verify some basic behavior of the 'to_bytes' and 'from_bytes' functions.
def _test_to_bytes_from_bytes():
    #   Python list.
    a = ['a', 'bcd', None, 1, 3.04, 'êƒ≈ß ', [1,2.0,"-3"], 'bcd', [1,2.0,"-3"], None]
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (tuple(map(str,a)) == tuple(map(lambda v: v.decode(), b))), f"\n{a}\n{b}"
    #   Numpy array of objects.
    a = np.asarray([['a', 'bcd', None], [1, 3.04, 'êƒ≈ß '], [1, 2.0, "-3"]], dtype=object)
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (tuple(map(str,a.flatten())) == tuple(map(lambda v: v.decode(), b))), f"\n{a}\n{b}"
    #   Numpy array of uint8.
    a = np.asarray([0, 4, 30, 255, 30, 4, 0], dtype="uint8")
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (tuple(a.flatten()) == tuple(b)), f"\n{a}\n{b}"
    #   Numpy array of int32.
    a = np.asarray([0, 4, 30, 255, 30, 4, 0], dtype="int32")
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (tuple(a.flatten()) == tuple(b)), f"\n{a}\n{b}"
    #   Numpy array of int64.
    a = np.asarray([0, 4, 30, 255, 30, 4, 0], dtype="int64") * 2**55
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (tuple(a.flatten()) == tuple(b)), f"\n{a}\n{b}"
    #   Numpy array of float32.
    a = np.asarray([float('inf'), 4.222, 30.178, -255.1830, float('nan'), 4.222, float('inf'), float('nan')], dtype="float32")
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (str(tuple(a.flatten())) == str(tuple(b))), f"\n{a}\n{b}"
    #   Numpy array of float64.
    a = np.asarray([float('inf'), 4.222, 30.178, -255.1830, float('nan'), 4.222, float('inf'), float('nan')], dtype="float64")
    output = to_bytes(a)
    b = from_bytes(*output)
    assert (str(tuple(a.flatten())) == str(tuple(b))), f"\n{a}\n{b}"

# Verify behavior of the 'unique' function.
def _test_unique():
    # Run a test, raise an error if it failes.
    def _run_test(a, test, fromlist=False):
        # Use the provided functions to generate a guess.
        u = unique(a)
        if fromlist: u = ByteArray.fromlist(u.tolist())
        i = to_int(test, u)
        # Generate a "ground truth" answer.
        if (u.info.value == b"object"):
            su = sorted(set(map(str, a)), key=lambda v: (len(v), v))
            si = np.asarray([su.index(str(v))+1 if str(v) in su else 0 for v in test])
        else:
            su = sorted(set(a.flatten()))
            si = np.asarray([su.index(v)+1 if v in su else 0 for v in test])
        if (not all(i == si)):
            print("a:    ", a, flush=True)
            print("test: ", test, flush=True)
            print()
            print("unique(a):    ", u, flush=True)
            print("to_int(a, u): ", i, flush=True)
            print()
            print("unique(a):    ", su, flush=True)
            print("to_int(a, u): ", si, flush=True)
            print()
            raise(ValueError(f"Expected the integer mappings to be the same."))

    np.random.seed(0)

    # Test with python list of objects.
    a = ['a', 'bcd', None, 1, 3.04, 'êƒ≈ß ', [1,2.0,"-3"], 'bcd', [1,2.0,"-3"], None]
    test = [a[3], a[6], -100.0] + a[1::2] + [10]
    _run_test(a, test)
    _run_test(a, test, fromlist=True)

    # Test with numpy array of uint8.
    a = np.arange(0, 256, 16).astype("uint8")
    np.random.shuffle(a)
    a = a[:8]
    test = np.arange(0, 256).astype("uint")
    _run_test(a, test)
    _run_test(a, test, fromlist=True)

    # Test with numpy array of ints
    a = np.arange(-2**30, 2**30, 2**27).astype("int32")
    np.random.shuffle(a)
    a = a[:8]
    test = np.arange(0, 128).astype("int32")
    _run_test(a, test)
    _run_test(a, test, fromlist=True)

    # Test with numpy array of longs
    a = np.arange(-2**61, 2**61, 2**58).astype("int64")
    np.random.shuffle(a)
    a = a[:8]
    test = np.arange(0, 128).astype("int64")
    _run_test(a, test)
    _run_test(a, test, fromlist=True)



# For testing purposes.
if __name__ == "__main__":
    # Run logic tests.
    _test_to_bytes_from_bytes()
    _test_unique()

    # Make sure we can force garbage collection.
    import gc
    # Ensure that profiling is enabled.
    from tlux.profiling import profile, inspect_memory, mem_use
    profiling = False
    if (profiling and (not hasattr(unique, "show_profile"))):
        unique = profile(unique)
    if (profiling and (not hasattr(to_int, "show_profile"))):
        to_int = profile(to_int)

    # Define a test array.
    array = np.asarray([['abc', 'de', 'f',  None, 'ghijk', 'f',  'de'],
                        [unique, 10,   1,   3.0,  [0,1],   (0,1), 10]], dtype=object)

    inspect_memory()

    # Test "unique".
    u = unique(array)
    print()
    print(u)
    truth = sorted(
        set(map(str, array.flatten().tolist())),
        key=lambda i: (len(i), i)
    )
    print(truth)
    print()

    assert (tuple(u) == tuple(truth)), f"ERROR: C-Library generated values do not match truth.\n  clib: {u}\n truth: {truth}\n"

    # Test "to_int".
    print(array)
    array[0,1] = 'NOPE' # to_int
    i = to_int(array, u)
    print()
    print(i)
    truth = np.asarray(
        [truth.index(str(v))+1 if str(v) in truth else 0
         for v in array.flatten()],
        dtype=int
    ).reshape(array.shape)
    print(truth)
    print()

    assert (tuple(i.tolist()) == tuple(truth.flatten().tolist())), f"ERROR: C-Library generated values do not match truth.\n  clib: {i.tolist()}\n truth: {truth.tolist()}\n"

    # Show initial profiles.
    if (hasattr(unique, "show_profile")): unique.show_profile()
    if (hasattr(to_int, "show_profile")): to_int.show_profile()
    print(f"mem_use: {mem_use():.2f}MB", flush=True)

    # Call the functions repeatedly and look for a memory leak.
    # @profile
    def testing(steps=10000):
        print("_"*70)
        print(f"Calling {steps} times..")
        for i in range(steps):
            mapping = unique(array)
            i_array = to_int(array, mapping)
            del mapping
            del i_array
        print("^"*70)
        import gc
        gc.collect()
    if (profiling and (not hasattr(testing, "show_profile"))):
        testing = profile(testing)

    if (profiling): inspect_memory()
    for i in range(5):
        testing()
        print(f"mem_use: {mem_use():.2f}MB", flush=True)
        if (hasattr(testing, "show_profile")): testing.show_profile()
        if (profiling): inspect_memory()




'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.

! Module for matrix multiplication (absolutely crucial for APOS speed).
! Includes routines for orthogonalization, computing the SVD, and
! radializing data matrices with the SVD.
'''

import os
import ctypes
import platform
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "axy_matrix_operations." + platform.machine() + ".so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-fPIC', '-shared', '-O3', '-lblas', '-llapack']
_ordered_dependencies = ['axy_matrix_operations.f90', 'axy_matrix_operations_c_wrapper.f90']
_symbol_files = []# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the prerequisite symbols for the compiled code.
for _ in _symbol_files:
    _ = ctypes.CDLL(os.path.join(_this_directory, _), mode=ctypes.RTLD_GLOBAL)
# Try to import the existing object. If that fails, recompile and then try.
try:
    # Check to see if the source files have been modified and a recompilation is needed.
    if (max(max([0]+[os.path.getmtime(os.path.realpath(os.path.join(_this_directory,_))) for _ in _symbol_files]),
            max([0]+[os.path.getmtime(os.path.realpath(os.path.join(_this_directory,_))) for _ in _ordered_dependencies]))
        > os.path.getmtime(_path_to_lib)):
        print()
        print("WARNING: Recompiling because the modification time of a source file is newer than the library.", flush=True)
        print()
        if os.path.exists(_path_to_lib):
            os.remove(_path_to_lib)
        raise NotImplementedError(f"The newest library code has not been compiled.")
    # Import the library.
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


class matrix_operations:
    ''''''

    # Declare 'pi'
    def get_pi(self):
        pi = ctypes.c_float()
        clib.matrix_operations_get_pi(ctypes.byref(pi))
        return pi.value
    def set_pi(self, pi):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    pi = property(get_pi, set_pi)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine GEMM
    
    def gemm(self, op_a, op_b, out_rows, out_cols, inner_dim, ab_mult, a, a_rows, b, b_rows, c_mult, c, c_rows):
        '''! Convenience wrapper routine for calling matrix multiply.
!
! TODO: For very large data, automatically batch the operations to reduce
!       the size of the temporary space that is needed.'''
        
        # Setting up "op_a"
        if (type(op_a) is not ctypes.c_char): op_a = ctypes.c_char(op_a)
        
        # Setting up "op_b"
        if (type(op_b) is not ctypes.c_char): op_b = ctypes.c_char(op_b)
        
        # Setting up "out_rows"
        if (type(out_rows) is not ctypes.c_int): out_rows = ctypes.c_int(out_rows)
        
        # Setting up "out_cols"
        if (type(out_cols) is not ctypes.c_int): out_cols = ctypes.c_int(out_cols)
        
        # Setting up "inner_dim"
        if (type(inner_dim) is not ctypes.c_int): inner_dim = ctypes.c_int(inner_dim)
        
        # Setting up "ab_mult"
        if (type(ab_mult) is not ctypes.c_float): ab_mult = ctypes.c_float(ab_mult)
        
        # Setting up "a"
        if ((not issubclass(type(a), numpy.ndarray)) or
            (not numpy.asarray(a).flags.f_contiguous) or
            (not (a.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'a' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            a = numpy.asarray(a, dtype=ctypes.c_float, order='F')
        a_dim_1 = ctypes.c_long(a.shape[0])
        a_dim_2 = ctypes.c_long(a.shape[1])
        
        # Setting up "a_rows"
        if (type(a_rows) is not ctypes.c_int): a_rows = ctypes.c_int(a_rows)
        
        # Setting up "b"
        if ((not issubclass(type(b), numpy.ndarray)) or
            (not numpy.asarray(b).flags.f_contiguous) or
            (not (b.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'b' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            b = numpy.asarray(b, dtype=ctypes.c_float, order='F')
        b_dim_1 = ctypes.c_long(b.shape[0])
        b_dim_2 = ctypes.c_long(b.shape[1])
        
        # Setting up "b_rows"
        if (type(b_rows) is not ctypes.c_int): b_rows = ctypes.c_int(b_rows)
        
        # Setting up "c_mult"
        if (type(c_mult) is not ctypes.c_float): c_mult = ctypes.c_float(c_mult)
        
        # Setting up "c"
        if ((not issubclass(type(c), numpy.ndarray)) or
            (not numpy.asarray(c).flags.f_contiguous) or
            (not (c.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'c' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            c = numpy.asarray(c, dtype=ctypes.c_float, order='F')
        c_dim_1 = ctypes.c_long(c.shape[0])
        c_dim_2 = ctypes.c_long(c.shape[1])
        
        # Setting up "c_rows"
        if (type(c_rows) is not ctypes.c_int): c_rows = ctypes.c_int(c_rows)
    
        # Call C-accessible Fortran wrapper.
        clib.c_gemm(ctypes.byref(op_a), ctypes.byref(op_b), ctypes.byref(out_rows), ctypes.byref(out_cols), ctypes.byref(inner_dim), ctypes.byref(ab_mult), ctypes.byref(a_dim_1), ctypes.byref(a_dim_2), ctypes.c_void_p(a.ctypes.data), ctypes.byref(a_rows), ctypes.byref(b_dim_1), ctypes.byref(b_dim_2), ctypes.c_void_p(b.ctypes.data), ctypes.byref(b_rows), ctypes.byref(c_mult), ctypes.byref(c_dim_1), ctypes.byref(c_dim_2), ctypes.c_void_p(c.ctypes.data), ctypes.byref(c_rows))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return c

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine SYRK
    
    def syrk(self, uplo, trans, n, k, alpha, a, lda, beta, c, ldc):
        '''! Convenience wrapper routine for calling symmetric matrix multiplication.'''
        
        # Setting up "uplo"
        if (type(uplo) is not ctypes.c_char): uplo = ctypes.c_char(uplo)
        
        # Setting up "trans"
        if (type(trans) is not ctypes.c_char): trans = ctypes.c_char(trans)
        
        # Setting up "n"
        if (type(n) is not ctypes.c_int): n = ctypes.c_int(n)
        
        # Setting up "k"
        if (type(k) is not ctypes.c_int): k = ctypes.c_int(k)
        
        # Setting up "alpha"
        if (type(alpha) is not ctypes.c_float): alpha = ctypes.c_float(alpha)
        
        # Setting up "a"
        if ((not issubclass(type(a), numpy.ndarray)) or
            (not numpy.asarray(a).flags.f_contiguous) or
            (not (a.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'a' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            a = numpy.asarray(a, dtype=ctypes.c_float, order='F')
        a_dim_1 = ctypes.c_long(a.shape[0])
        a_dim_2 = ctypes.c_long(a.shape[1])
        
        # Setting up "lda"
        if (type(lda) is not ctypes.c_int): lda = ctypes.c_int(lda)
        
        # Setting up "beta"
        if (type(beta) is not ctypes.c_float): beta = ctypes.c_float(beta)
        
        # Setting up "c"
        if ((not issubclass(type(c), numpy.ndarray)) or
            (not numpy.asarray(c).flags.f_contiguous) or
            (not (c.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'c' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            c = numpy.asarray(c, dtype=ctypes.c_float, order='F')
        c_dim_1 = ctypes.c_long(c.shape[0])
        c_dim_2 = ctypes.c_long(c.shape[1])
        
        # Setting up "ldc"
        if (type(ldc) is not ctypes.c_int): ldc = ctypes.c_int(ldc)
    
        # Call C-accessible Fortran wrapper.
        clib.c_syrk(ctypes.byref(uplo), ctypes.byref(trans), ctypes.byref(n), ctypes.byref(k), ctypes.byref(alpha), ctypes.byref(a_dim_1), ctypes.byref(a_dim_2), ctypes.c_void_p(a.ctypes.data), ctypes.byref(lda), ctypes.byref(beta), ctypes.byref(c_dim_1), ctypes.byref(c_dim_2), ctypes.c_void_p(c.ctypes.data), ctypes.byref(ldc))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return c

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine ORTHOGONALIZE
    
    def orthogonalize(self, a, lengths, rank=None, order=None, multipliers=None):
        '''! Orthogonalize and normalize column vectors of A with pivoting.'''
        
        # Setting up "a"
        if ((not issubclass(type(a), numpy.ndarray)) or
            (not numpy.asarray(a).flags.f_contiguous) or
            (not (a.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'a' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            a = numpy.asarray(a, dtype=ctypes.c_float, order='F')
        a_dim_1 = ctypes.c_long(a.shape[0])
        a_dim_2 = ctypes.c_long(a.shape[1])
        
        # Setting up "lengths"
        if ((not issubclass(type(lengths), numpy.ndarray)) or
            (not numpy.asarray(lengths).flags.f_contiguous) or
            (not (lengths.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'lengths' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            lengths = numpy.asarray(lengths, dtype=ctypes.c_float, order='F')
        lengths_dim_1 = ctypes.c_long(lengths.shape[0])
        
        # Setting up "rank"
        rank_present = ctypes.c_bool(True)
        if (rank is None):
            rank_present = ctypes.c_bool(False)
            rank = ctypes.c_int()
        else:
            rank = ctypes.c_int(rank)
        
        # Setting up "order"
        order_present = ctypes.c_bool(True)
        if (order is None):
            order_present = ctypes.c_bool(False)
            order = numpy.zeros(shape=(1), dtype=ctypes.c_int, order='F')
        elif (type(order) == bool) and (order):
            order = numpy.zeros(shape=(1), dtype=ctypes.c_int, order='F')
        elif ((not issubclass(type(order), numpy.ndarray)) or
              (not numpy.asarray(order).flags.f_contiguous) or
              (not (order.dtype == numpy.dtype(ctypes.c_int)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_int' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_int, order='F')
        if (order_present):
            order_dim_1 = ctypes.c_long(order.shape[0])
        else:
            order_dim_1 = ctypes.c_long()
        
        # Setting up "multipliers"
        multipliers_present = ctypes.c_bool(True)
        if (multipliers is None):
            multipliers_present = ctypes.c_bool(False)
            multipliers = numpy.zeros(shape=(1,1), dtype=ctypes.c_float, order='F')
        elif (type(multipliers) == bool) and (multipliers):
            multipliers = numpy.zeros(shape=(1,1), dtype=ctypes.c_float, order='F')
        elif ((not issubclass(type(multipliers), numpy.ndarray)) or
              (not numpy.asarray(multipliers).flags.f_contiguous) or
              (not (multipliers.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'multipliers' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            multipliers = numpy.asarray(multipliers, dtype=ctypes.c_float, order='F')
        if (multipliers_present):
            multipliers_dim_1 = ctypes.c_long(multipliers.shape[0])
            multipliers_dim_2 = ctypes.c_long(multipliers.shape[1])
        else:
            multipliers_dim_1 = ctypes.c_long()
            multipliers_dim_2 = ctypes.c_long()
    
        # Call C-accessible Fortran wrapper.
        clib.c_orthogonalize(ctypes.byref(a_dim_1), ctypes.byref(a_dim_2), ctypes.c_void_p(a.ctypes.data), ctypes.byref(lengths_dim_1), ctypes.c_void_p(lengths.ctypes.data), ctypes.byref(rank_present), ctypes.byref(rank), ctypes.byref(order_present), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(multipliers_present), ctypes.byref(multipliers_dim_1), ctypes.byref(multipliers_dim_2), ctypes.c_void_p(multipliers.ctypes.data))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return a, lengths, (rank.value if rank_present else None), (order if order_present else None), (multipliers if multipliers_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine SVD
    
    def svd(self, a, s, vt, rank=None, steps=None, bias=None):
        '''! Compute the singular values and right singular vectors for matrix A.
!
! TODO: Add an INFO flag that can be used to raise errors.
! TODO: Move allocation to be allowed as input.'''
        
        # Setting up "a"
        if ((not issubclass(type(a), numpy.ndarray)) or
            (not numpy.asarray(a).flags.f_contiguous) or
            (not (a.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'a' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            a = numpy.asarray(a, dtype=ctypes.c_float, order='F')
        a_dim_1 = ctypes.c_long(a.shape[0])
        a_dim_2 = ctypes.c_long(a.shape[1])
        
        # Setting up "s"
        if ((not issubclass(type(s), numpy.ndarray)) or
            (not numpy.asarray(s).flags.f_contiguous) or
            (not (s.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 's' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            s = numpy.asarray(s, dtype=ctypes.c_float, order='F')
        s_dim_1 = ctypes.c_long(s.shape[0])
        
        # Setting up "vt"
        if ((not issubclass(type(vt), numpy.ndarray)) or
            (not numpy.asarray(vt).flags.f_contiguous) or
            (not (vt.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'vt' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            vt = numpy.asarray(vt, dtype=ctypes.c_float, order='F')
        vt_dim_1 = ctypes.c_long(vt.shape[0])
        vt_dim_2 = ctypes.c_long(vt.shape[1])
        
        # Setting up "rank"
        rank_present = ctypes.c_bool(True)
        if (rank is None):
            rank_present = ctypes.c_bool(False)
            rank = ctypes.c_int()
        else:
            rank = ctypes.c_int(rank)
        
        # Setting up "steps"
        steps_present = ctypes.c_bool(True)
        if (steps is None):
            steps_present = ctypes.c_bool(False)
            steps = ctypes.c_int()
        else:
            steps = ctypes.c_int(steps)
        if (type(steps) is not ctypes.c_int): steps = ctypes.c_int(steps)
        
        # Setting up "bias"
        bias_present = ctypes.c_bool(True)
        if (bias is None):
            bias_present = ctypes.c_bool(False)
            bias = ctypes.c_float()
        else:
            bias = ctypes.c_float(bias)
        if (type(bias) is not ctypes.c_float): bias = ctypes.c_float(bias)
    
        # Call C-accessible Fortran wrapper.
        clib.c_svd(ctypes.byref(a_dim_1), ctypes.byref(a_dim_2), ctypes.c_void_p(a.ctypes.data), ctypes.byref(s_dim_1), ctypes.c_void_p(s.ctypes.data), ctypes.byref(vt_dim_1), ctypes.byref(vt_dim_2), ctypes.c_void_p(vt.ctypes.data), ctypes.byref(rank_present), ctypes.byref(rank), ctypes.byref(steps_present), ctypes.byref(steps), ctypes.byref(bias_present), ctypes.byref(bias))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return s, vt, (rank.value if rank_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine RADIALIZE
    
    def radialize(self, x, shift, vecs, inverse=None, maxbound=None, max_to_flatten=None, max_to_square=None, svd_steps=None):
        '''! If there are at least as many data points as dimension, then
! compute the principal components and rescale the data by
! projecting onto those and rescaling so that each component has
! identical singular values (this makes the data more "radially
! symmetric").
!
!   X -- The real data matrix, column vectors (a "point" is a column).
!
!   SHIFT -- The real vector that is added to X in order to center it about the origin.
!
!   VECS -- The approximate principal components of the data, these are the orthogonal
!           column vectors about which the data in X is rotated. Notably, they are NOT
!           necessarily unit length, because they are rescaled to achieve the desired
!           properties in X (which property is determined by MAXBOUND setting).
!
!   INVERSE -- Optional output, the inverse of VECS.
!
!   MAX_TO_FLATTEN -- Optional input, the integer maximum number of components to flatten
!                     (rescale by their singular value) while the rest simply get divided
!                     by the first (maximum) singular value.
!
!   MAXBOUND -- Optional input, TRUE to normalize by the first (maximum) singular value,
!               defaults to FALSE, to normalize so that the 2-norm of the data matrix is 1.
!
!   MAX_TO_SQUARE -- Optional, the integer maximum number of points (second component of X)
!                    that will be considered for the routine (to bound compute for large data).
!
!   SVD_STEPS -- Optional input, the integer number of power-iterations to take when estimating
!                the principal components of the data with the SVD of the covariance matrix.'''
        
        # Setting up "x"
        if ((not issubclass(type(x), numpy.ndarray)) or
            (not numpy.asarray(x).flags.f_contiguous) or
            (not (x.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'x' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            x = numpy.asarray(x, dtype=ctypes.c_float, order='F')
        x_dim_1 = ctypes.c_long(x.shape[0])
        x_dim_2 = ctypes.c_long(x.shape[1])
        
        # Setting up "shift"
        if ((not issubclass(type(shift), numpy.ndarray)) or
            (not numpy.asarray(shift).flags.f_contiguous) or
            (not (shift.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'shift' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            shift = numpy.asarray(shift, dtype=ctypes.c_float, order='F')
        shift_dim_1 = ctypes.c_long(shift.shape[0])
        
        # Setting up "vecs"
        if ((not issubclass(type(vecs), numpy.ndarray)) or
            (not numpy.asarray(vecs).flags.f_contiguous) or
            (not (vecs.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'vecs' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            vecs = numpy.asarray(vecs, dtype=ctypes.c_float, order='F')
        vecs_dim_1 = ctypes.c_long(vecs.shape[0])
        vecs_dim_2 = ctypes.c_long(vecs.shape[1])
        
        # Setting up "inverse"
        inverse_present = ctypes.c_bool(True)
        if (inverse is None):
            inverse_present = ctypes.c_bool(False)
            inverse = numpy.zeros(shape=(1,1), dtype=ctypes.c_float, order='F')
        elif (type(inverse) == bool) and (inverse):
            inverse = numpy.zeros(shape=(1,1), dtype=ctypes.c_float, order='F')
        elif ((not issubclass(type(inverse), numpy.ndarray)) or
              (not numpy.asarray(inverse).flags.f_contiguous) or
              (not (inverse.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'inverse' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            inverse = numpy.asarray(inverse, dtype=ctypes.c_float, order='F')
        if (inverse_present):
            inverse_dim_1 = ctypes.c_long(inverse.shape[0])
            inverse_dim_2 = ctypes.c_long(inverse.shape[1])
        else:
            inverse_dim_1 = ctypes.c_long()
            inverse_dim_2 = ctypes.c_long()
        
        # Setting up "maxbound"
        maxbound_present = ctypes.c_bool(True)
        if (maxbound is None):
            maxbound_present = ctypes.c_bool(False)
            maxbound = ctypes.c_int()
        else:
            maxbound = ctypes.c_int(maxbound)
        if (type(maxbound) is not ctypes.c_int): maxbound = ctypes.c_int(maxbound)
        
        # Setting up "max_to_flatten"
        max_to_flatten_present = ctypes.c_bool(True)
        if (max_to_flatten is None):
            max_to_flatten_present = ctypes.c_bool(False)
            max_to_flatten = ctypes.c_int()
        else:
            max_to_flatten = ctypes.c_int(max_to_flatten)
        if (type(max_to_flatten) is not ctypes.c_int): max_to_flatten = ctypes.c_int(max_to_flatten)
        
        # Setting up "max_to_square"
        max_to_square_present = ctypes.c_bool(True)
        if (max_to_square is None):
            max_to_square_present = ctypes.c_bool(False)
            max_to_square = ctypes.c_long()
        else:
            max_to_square = ctypes.c_long(max_to_square)
        if (type(max_to_square) is not ctypes.c_long): max_to_square = ctypes.c_long(max_to_square)
        
        # Setting up "svd_steps"
        svd_steps_present = ctypes.c_bool(True)
        if (svd_steps is None):
            svd_steps_present = ctypes.c_bool(False)
            svd_steps = ctypes.c_int()
        else:
            svd_steps = ctypes.c_int(svd_steps)
        if (type(svd_steps) is not ctypes.c_int): svd_steps = ctypes.c_int(svd_steps)
    
        # Call C-accessible Fortran wrapper.
        clib.c_radialize(ctypes.byref(x_dim_1), ctypes.byref(x_dim_2), ctypes.c_void_p(x.ctypes.data), ctypes.byref(shift_dim_1), ctypes.c_void_p(shift.ctypes.data), ctypes.byref(vecs_dim_1), ctypes.byref(vecs_dim_2), ctypes.c_void_p(vecs.ctypes.data), ctypes.byref(inverse_present), ctypes.byref(inverse_dim_1), ctypes.byref(inverse_dim_2), ctypes.c_void_p(inverse.ctypes.data), ctypes.byref(maxbound_present), ctypes.byref(maxbound), ctypes.byref(max_to_flatten_present), ctypes.byref(max_to_flatten), ctypes.byref(max_to_square_present), ctypes.byref(max_to_square), ctypes.byref(svd_steps_present), ctypes.byref(svd_steps))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return x, shift, vecs, (inverse if inverse_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine LEAST_SQUARES
    
    def least_squares(self, trans, a, b, x):
        '''! Perform least squares with LAPACK.
!
!   A is column vectors (of points) if TRANS='T', and row vectors
!     (of points) if TRANS='N'.
!   B must be COLUMN VECTORS of fit output (1 row = 1 point).
!   X always has a first dimension that is nonpoint axis size of A,
!     and the second dimension is determined by B's columns (or rank),
!     and if X is smaller then B is reduced to its principal components.'''
        
        # Setting up "trans"
        if (type(trans) is not ctypes.c_char): trans = ctypes.c_char(trans)
        
        # Setting up "a"
        if ((not issubclass(type(a), numpy.ndarray)) or
            (not numpy.asarray(a).flags.f_contiguous) or
            (not (a.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'a' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            a = numpy.asarray(a, dtype=ctypes.c_float, order='F')
        a_dim_1 = ctypes.c_long(a.shape[0])
        a_dim_2 = ctypes.c_long(a.shape[1])
        
        # Setting up "b"
        if ((not issubclass(type(b), numpy.ndarray)) or
            (not numpy.asarray(b).flags.f_contiguous) or
            (not (b.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'b' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            b = numpy.asarray(b, dtype=ctypes.c_float, order='F')
        b_dim_1 = ctypes.c_long(b.shape[0])
        b_dim_2 = ctypes.c_long(b.shape[1])
        
        # Setting up "x"
        if ((not issubclass(type(x), numpy.ndarray)) or
            (not numpy.asarray(x).flags.f_contiguous) or
            (not (x.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'x' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            x = numpy.asarray(x, dtype=ctypes.c_float, order='F')
        x_dim_1 = ctypes.c_long(x.shape[0])
        x_dim_2 = ctypes.c_long(x.shape[1])
    
        # Call C-accessible Fortran wrapper.
        clib.c_least_squares(ctypes.byref(trans), ctypes.byref(a_dim_1), ctypes.byref(a_dim_2), ctypes.c_void_p(a.ctypes.data), ctypes.byref(b_dim_1), ctypes.byref(b_dim_2), ctypes.c_void_p(b.ctypes.data), ctypes.byref(x_dim_1), ctypes.byref(x_dim_2), ctypes.c_void_p(x.ctypes.data))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return a, b, x

matrix_operations = matrix_operations()




# The index-to-pair pattern matches the following.
# 
#   2 elements -> (1,1) (1,2) (2,1) (2,2)
#                   1     1     1     2
#                  1     1     2     2
#                    1     2     1     2
#   3 elements -> (1,1) (1,2) (2,1) (1,3) (3,1) (2,2) (2,3) (3,2) (3,3)
#                   1     1     1     1     1     2     2     2     3
#                  1     1     2     1     3     2     2     3     3
#                    1     2     1     3     1     2     3     2     3
#   4 elements -> (1,1) (1,2) (2,1) (1,3) (3,1) (1,4) (4,1) (2,2) (2,3) (3,2) (2,4) (4,2) (3,3) (3,4) (4,3) (4,4)
#                   1     1     1     1     1     1     1     2     2     2     2     2     3     3     3     4
#                  1     1     2     1     3     1     4     2     2     3     2     4     3     3     4     4
#                    1     2     1     3     1     4     1     2     3     2     4     2     3     4     3     4
# 
# From the back, the number of elements such that the min is i goes:
#    1  3  5  7 ... (1 + 2 <num-others>)
# 
# so given there are n unique elements, the number of elements with min(i) forward is:
#   (1+2n)  (1+2(n-1))  ...  3  1
# 
#  n - (i-n) / 2 - 1


def index_to_pair(index_of_pair, num_elements):
    index_from_back = num_elements**2 - index_of_pair
    group_from_back = int(index_from_back**(1/2))
    group = num_elements - group_from_back
    remaining_in_group = index_from_back - group_from_back**2
    group_size = 2*group_from_back + 1
    index_in_group = group_size - remaining_in_group
    other = group + index_in_group // 2
    if (index_in_group % 2 == 0):
        return (group, other)
    else:
        return (other, group)


def pair_to_index(pair, num_elements):
    group = min(pair)
    index_in_group = 2 * (max(pair) - group)
    if (pair[0] < pair[1]):
        index_in_group -= 1
    group_from_back = 1 + num_elements - group
    group_start_position = num_elements**2 - group_from_back**2 + 1
    index = group_start_position + index_in_group
    return index


# Abstracted version of the index to pair function.
def i2p(i, n):
    # b = n**2 - i
    # c = int(b**(1/2))
    # g = n - c
    # r = b - c**2
    # s = 2*c + 1
    # j = s - r
    #   = (2c + 1) - (b - c2)
    #   = 2c + 1 - b - c2
    #   = c2 + 2c + 1 - b
    #   = (c+1)2 - n2 + i
    # o = g + j // 2
    # 
    n2 = n**2
    c = int((n2 - i)**(1/2))
    g = n - c
    j = (c+1)**2 - n2 + i
    o = g + j // 2
    if (j % 2 == 0):
        return (g, o)
    else:
        return (o, g)

# Abstracted version of the pair to index function.
def p2i(p, n):
    p1, p2 = p
    # 
    # g = min(p1, p2)
    # o = max(p1, p2)
    # j = 2 * (o - g)
    # if (p[0] < p[1]):
    #     j -= 1
    # c = 1 + n - g
    # s = n**2 - c**2 + 1
    # i = s + j
    # 
    # if (p1 < p2):
    #     g = p1
    #     o = p2
    #     j = 2 * (o - g) - 1
    #     c = 1 + n - g
    #     s = n**2 - c**2 + 1
    #     i = s + j
    # else:
    #     g = p2
    #     o = p1
    #     j = 2 * (o - g)
    #     c = 1 + n - g
    #     s = n**2 - c**2 + 1
    #     i = s + j
    # 
    if (p1 < p2):
        i = 2*(n*(p1 - 1) + p2) - p1**2 - 1
    else:
        i = 2*(n*(p2 - 1) + p1) - p2**2
    return i


print()

# Generate example.
n = 4
print("Pretty")
for i in range(1, n**2+1):
    print(f'{i:3d} -> {index_to_pair(i, n)} -> {pair_to_index(index_to_pair(i,n),n)}')
print()
print("Abstracted")
for i in range(1, n**2+1):
    print(f'{i:3d} -> {index_to_pair(i, n)} -> {p2i(i2p(i,n),n)}')
print()

# Matrix operations module.
import fmodpy
random = fmodpy.fimport("../axy_random.f90", dependencies=["pcg32.f90"], name="test_axy_random").random
print("Fortran")
for i in range(1, n**2+1):
    print(f'{i:3d} -> {random.index_to_pair(n, i)} -> {random.pair_to_index(n, *random.index_to_pair(n,i))}')
